from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np
from scipy.optimize import root

from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions
    from chemex.typing import ArrayFloat

NAME = "2st_binding"

TPL = ("temperature", "p_total", "l_total")


def calculate_residuals(
    concentrations: ArrayFloat,
    p_total: float,
    l_total: float,
    kd: float,
) -> ArrayFloat:
    p_free, l_free, pl = concentrations
    return np.array(
        [
            l_total - l_free - pl,
            p_total - p_free - pl,
            kd * pl - p_free * l_free,
        ],
    )


@lru_cache(maxsize=100)
def calculate_concentrations(
    p_total: float,
    l_total: float,
    kd: float,
) -> dict[str, float]:
    concentrations_start = (p_total, l_total, 0.0)
    results = root(
        calculate_residuals,
        concentrations_start,
        args=(p_total, l_total, kd),
    )
    p_free, l_free, pl = results["x"]
    return {"p_free": p_free, "l_free": l_free, "pl": pl}


def make_settings_2st_binding(conditions: Conditions) -> dict[str, ParamLocalSetting]:
    p_total = conditions.p_total
    l_total = conditions.l_total
    if p_total is None or l_total is None:
        msg = "'p_total' and 'l_total' must be specified to use the '2st_binding' model"
        raise ValueError(msg)
    return {
        "koff": ParamLocalSetting(
            name_setting=NameSetting("koff", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "kd": ParamLocalSetting(
            name_setting=NameSetting("kd", "", ("temperature",)),
            value=1e-3,
            min=0.0,
            vary=True,
        ),
        "kon": ParamLocalSetting(
            name_setting=NameSetting("kon", "", ("temperature",)),
            expr="{koff} / max({kd}, 1e-100)",
        ),
        "p_free": ParamLocalSetting(
            name_setting=NameSetting("p_free", "", TPL),
            expr=f"calc_conc({p_total}, {l_total}, {{kd}})['p_free']",
        ),
        "l_free": ParamLocalSetting(
            name_setting=NameSetting("l_free", "", TPL),
            expr=f"calc_conc({p_total}, {l_total}, {{kd}})['l_free']",
        ),
        "pl": ParamLocalSetting(
            name_setting=NameSetting("pl", "", TPL),
            expr=f"calc_conc({p_total}, {l_total}, {{kd}})['pl']",
        ),
        "kab": ParamLocalSetting(
            name_setting=NameSetting("kab", "", TPL),
            expr="{kon} * {l_free}",
        ),
        "kba": ParamLocalSetting(
            name_setting=NameSetting("kba", "", TPL),
            expr="{koff}",
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TPL),
            expr=f"{{p_free}} / {p_total}",
        ),
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "", TPL),
            expr=f"{{pl}} / {p_total}",
        ),
    }


def register() -> None:
    model_factory.register(name=NAME, setting_maker=make_settings_2st_binding)
    user_functions = {"calc_conc": calculate_concentrations}
    user_function_registry.register(name=NAME, user_functions=user_functions)



"""ChemEx is an analysis program for chemical exchange detected by NMR.

It is designed to take almost any kind of NMR data to aid the analysis,
but the principle techniques are CPMG relaxation dispersion and Chemical
Exchange Saturation Transfer.

"""
from importlib import metadata

__version__ = metadata.version(__name__)



from __future__ import annotations

import re
from dataclasses import dataclass, field
from functools import cached_property
from re import Pattern
from typing import TYPE_CHECKING, TypeVar

from rapidfuzz.process import extractOne

from chemex.configuration.conditions import Conditions
from chemex.parameters.spin_system import SpinSystem

if TYPE_CHECKING:
    from collections.abc import Hashable, Iterable

Self = TypeVar("Self", bound="ParamName")

_EXPAND = {"-": "_", "+": "_", ".": "_"}

_FLOAT = r"[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?"

_RE_SECTION = re.compile(
    rf"""
        (^\s*(?P<name>\w+)) |
        (NUC\s*->\s*(?P<spin_system>(\w|-)+)) |
        (T\s*->s*(?P<temperature>{_FLOAT})) |
        (B0\s*->\s*(?P<h_larmor_frq>{_FLOAT})) |
        (\[P]\s*->\s*(?P<p_total>{_FLOAT})) |
        (\[L]\s*->\s*(?P<l_total>{_FLOAT})) |
        (D2O\s*->\s*(?P<d2o>{_FLOAT}))
    """,
    re.IGNORECASE | re.VERBOSE,
)


def _parse(re_to_match: Pattern[str], text: str) -> dict[str, str]:
    return {
        key: value
        for match in re_to_match.finditer(text)
        for key, value in match.groupdict().items()
        if value is not None
    }


def _multireplace(string: str, replacements: dict[str, str]) -> str:
    """Given a string and a replacement map, it returns the replaced string.

    :param str string: string to execute replacements on
    :param dict replacements: replacement dictionary {value to find: value to replace}
    :rtype: str

    """
    # Place longer ones first to keep shorter substrings from matching where the longer
    # ones should take place. For instance given the replacements
    # {'ab': 'AB', 'abc': 'ABC'} against the string 'hey abc', it should produce
    # 'hey ABC' and not 'hey ABc'
    substrings = sorted(replacements, key=len, reverse=True)

    # Create a big OR regex that matches any of the substrings to replace
    regexp = re.compile("|".join(re.escape(substring) for substring in substrings))

    # For each match, look up the new string in the replacements
    return regexp.sub(lambda match: replacements[match.group(0)], string)


def _expand(string: str) -> str:
    return _multireplace(string, _EXPAND)


@dataclass(order=True, unsafe_hash=True)
class ParamName:
    name: str = ""
    spin_system: SpinSystem = field(default_factory=SpinSystem)
    conditions: Conditions = field(default_factory=Conditions.model_construct)
    search_keys: set[Hashable] = field(init=False, compare=False)

    def __post_init__(self) -> None:
        self.name = self.name.strip().upper()
        self.conditions = self.conditions.rounded()
        self.search_keys = (
            self.spin_system.search_keys | self.conditions.search_keys | {self.name}
        )

    @classmethod
    def from_section(cls: type[Self], section: str = "") -> Self:
        parsed = _parse(_RE_SECTION, section.strip(' []"'))
        name = parsed.get("name")
        if name is None:
            name = ""
        spin_system = SpinSystem(parsed.get("spin_system"))
        conditions = Conditions.model_validate(parsed)
        return cls(name, spin_system, conditions)

    @cached_property
    def section(self) -> str:
        parts = (self.name, self.conditions.section)
        return ", ".join(parts).strip(", ").upper()

    @cached_property
    def section_res(self) -> str:
        parts = [self.name]
        if self.spin_system:
            parts.append(f"NUC->{self.spin_system}")
        parts.append(self.conditions.section)
        return ", ".join(parts).strip(", ").upper()

    @cached_property
    def folder(self) -> str:
        parts = (self.name, str(self.spin_system), self.conditions.folder)
        return "_".join(parts).strip("_").upper()

    @cached_property
    def id_(self) -> str:
        return f"__{_expand(self.folder)}"

    def match(self: Self, other: Self) -> bool:
        if self.name and self.name not in other.name:
            return False
        if not self.spin_system.match(other.spin_system):
            return False
        if not self.conditions.match(other.conditions):
            return False
        return True

    def get_closest_id(self, ids: Iterable[str]) -> str:
        best_match, _, _ = extractOne(self.id_, ids)
        return best_match

    def __and__(self: Self, other: object) -> Self:
        if not isinstance(other, type(self)):
            return NotImplemented
        name = self.name if self.name == other.name else ""
        spin_system = self.spin_system & other.spin_system
        conditions = self.conditions & other.conditions
        return type(self)(name, spin_system, conditions)

    def __bool__(self) -> bool:
        return bool(self.name) or bool(self.spin_system) or bool(self.conditions)

    def __str__(self) -> str:
        return f"[{self.section_res}]"



"""The module provides functions for estimating the noise variance in data."""
from __future__ import annotations

import contextlib
from collections import defaultdict
from typing import TYPE_CHECKING, Any

import numpy as np
from scipy import interpolate, signal, stats
from scipy.linalg import norm

if TYPE_CHECKING:
    from collections.abc import Callable

    from chemex.containers.data import Data


def _variance_from_duplicates(data: Data) -> float:
    """Estimates variance from duplicate data points in a dataset.

    This function calculates the variance using the pooled standard deviation method.
    It groups the data points by their metadata and computes the variance for each
    group. If no duplicates are found, the average of the experimental error values is
    returned.

    Args:
        data (Data): The data containing experimental values and associated metadata.

    Returns:
        float: The estimated variance of the dataset.

    Reference:
        IUPAC Gold Book definition of pooled standard deviation.
        URL: https://goldbook.iupac.org/html/P/P04758.html
    """
    groups: defaultdict[Any, list[float]] = defaultdict(list)
    for x, y in zip(data.metadata, data.exp, strict=True):
        groups[x].append(y)
    variances: list[float] = []
    weights: list[float] = []
    for group in groups.values():
        group_size = len(group)
        if group_size > 1:
            variances.append(np.var(group, ddof=1))
            weights.append(group_size - 1)
    if not variances:
        return float(np.mean(data.err))
    return float(np.average(variances, weights=weights))


def _variance_from_scatter(data: Data) -> float:
    """Estimates the uncertainty in the CEST profile or similar data.

    The function uses numerical methods to estimate the variance. It applies finite
    difference approximation and interpolates the sorted experimental data to estimate
    the standard deviation at different percentages of the data. The final variance is
    calculated as the median of these estimates, adjusted for the size of the data.

    Args:
        data (Data): The data containing experimental values.

    Returns:
        float: The estimated variance of the dataset.

    Reference:
        MATLAB Central File Exchange, 'EstimateNoise' function.
        URL: https://www.mathworks.com/matlabcentral/fileexchange/16683-estimatenoise
    """
    fda = [
        [1, -1],
        [1, -2, 1],
        [1, -3, 3, -1],
        [1, -4, 6, -4, 1],
        [1, -5, 10, -10, 5, -1],
        [1, -6, 15, -20, 15, -6, 1],
    ]
    fda = [np.array(a_fda) / norm(a_fda) for a_fda in fda]
    percents = np.array([0.05, *list(np.arange(0.1, 0.4, 0.025))])
    percent_points = stats.norm.ppf(1.0 - percents)
    sigma_est = []
    for fdai in fda:
        noisedata = sorted(signal.convolve(data.exp, fdai, mode="valid"))
        ntrim = len(noisedata)
        if ntrim <= 1:
            continue
        xaxis = (0.5 + np.arange(1, ntrim + 1)) / (ntrim + 0.5)
        sigmas: list[float] = []
        function = interpolate.interp1d(xaxis, noisedata, "linear")
        for a_perc, a_z in zip(percents, percent_points, strict=True):
            with contextlib.suppress(ValueError):
                val = (function(1.0 - a_perc) - function(a_perc)) / (2.0 * a_z)
                sigmas.append(val)
        sigma_est.append(np.median(sigmas))
    variance = np.median(sigma_est) ** 2 / (
        1.0 + 15.0 * (data.exp.size + 1.225) ** -1.245
    )
    return np.max([variance, 1e-8])


estimate_noise_variance: dict[str, Callable[[Data], float]] = {
    "scatter": _variance_from_scatter,
    "duplicates": _variance_from_duplicates,
}



from __future__ import annotations

from dataclasses import dataclass
from functools import reduce
from typing import TYPE_CHECKING, Literal

import numpy as np

from chemex.configuration.data import RelaxationDataSettings
from chemex.configuration.experiment import (
    CpmgSettingsEvenNcycs,
    ExperimentConfig,
    ToBeFitted,
)
from chemex.containers.dataset import load_relaxation_dataset
from chemex.experiments.factories import Creators, factories
from chemex.filterers import PlanesFilterer
from chemex.nmr.basis import Basis
from chemex.nmr.liouvillian import LiouvillianIS
from chemex.nmr.spectrometer import Spectrometer
from chemex.plotters.cpmg import CpmgPlotter
from chemex.printers.data import CpmgPrinter

if TYPE_CHECKING:
    from chemex.containers.data import Data
    from chemex.parameters.spin_system import SpinSystem
    from chemex.typing import ArrayBool, ArrayFloat, ArrayInt


EXPERIMENT_NAME = "cpmg_hn_dq_zq"

NU_CPMG_LIMIT_1 = 51.0
NU_CPMG_LIMIT_2 = 255.0


class CpmgHNDqZqSettings(CpmgSettingsEvenNcycs):
    name: Literal["cpmg_hn_dq_zq"]
    time_t2: float
    carrier_h: float
    carrier_n: float
    pw90_h: float
    pw90_n: float
    dq_flg: bool
    observed_state: Literal["a", "b", "c", "d"] = "a"

    @property
    def detection(self) -> str:
        if self.dq_flg:
            return f"[2ixsx_{self.observed_state}] - [2iysy_{self.observed_state}]"
        return f"[2ixsx_{self.observed_state}] + [2iysy_{self.observed_state}]"


class CpmgHNDqZqConfig(ExperimentConfig[CpmgHNDqZqSettings, RelaxationDataSettings]):
    @property
    def to_be_fitted(self) -> ToBeFitted:
        state = self.experiment.observed_state
        return ToBeFitted(
            rates=[f"r2mq_is_{state}", f"mu_is_{state}"],
            model_free=[f"tauc_{state}", f"s2_{state}"],
        )


def build_spectrometer(
    config: CpmgHNDqZqConfig,
    spin_system: SpinSystem,
) -> Spectrometer:
    settings = config.experiment
    conditions = config.conditions

    basis = Basis(type="ixyzsxyz", spin_system="nh")
    liouvillian = LiouvillianIS(spin_system, basis, conditions)
    spectrometer = Spectrometer(liouvillian)

    spectrometer.carrier_i = settings.carrier_n
    spectrometer.carrier_s = settings.carrier_h
    spectrometer.b1_i = 1 / (4.0 * settings.pw90_n)
    spectrometer.b1_s = 1 / (4.0 * settings.pw90_h)
    spectrometer.detection = settings.detection

    return spectrometer


@dataclass
class CpmgHNDqZqSequence:
    settings: CpmgHNDqZqSettings

    def _get_tau_cps(self, ncycs: ArrayFloat) -> dict[float, float]:
        ncycs_no_ref = ncycs[ncycs > 0]
        return dict(
            zip(
                ncycs_no_ref,
                self.settings.time_t2 / (4.0 * ncycs_no_ref)
                - 7.0 / 3.0 * self.settings.pw90_n,
                strict=True,
            ),
        )

    def _get_phases(self, ncyc: float) -> tuple[ArrayInt, ArrayInt]:
        nu_cpmg = ncyc / self.settings.time_t2
        if nu_cpmg < NU_CPMG_LIMIT_1:
            cp_phases1 = [0, 1, 0, 1]
            cp_phases2 = [1, 0, 1, 0]
        elif nu_cpmg < NU_CPMG_LIMIT_2:
            cp_phases1 = [0]
            cp_phases2 = [1]
        else:
            cp_phases1 = [0, 1, 0, 1, 1, 0, 1, 0]
            cp_phases2 = [1, 0, 1, 0, 0, 1, 0, 1]
        indexes = np.arange(2 * int(ncyc))
        phases1 = np.take(cp_phases1, np.flip(indexes), mode="wrap")
        phases2 = np.take(cp_phases2, np.flip(indexes), mode="wrap")
        return phases1, phases2

    def calculate(self, spectrometer: Spectrometer, data: Data) -> ArrayFloat:
        ncycs = data.metadata

        # Calculation of the spectrometers corresponding to all the delays
        tau_cps = self._get_tau_cps(ncycs)
        ncyc_list, delay_list = zip(*tau_cps.items(), strict=True)
        d_cp = dict(zip(ncyc_list, spectrometer.delays(delay_list), strict=True))

        # Calculation of the spectrometers corresponding to all the pulses
        p9024090_1 = spectrometer.p9024090_nh_1[[0, 1], [0, 1]]
        p9024090_2 = spectrometer.p9024090_nh_2[[0, 1], [0, 1]]

        # Getting the starting magnetization
        start = spectrometer.get_start_magnetization(["2ixsx"])

        # Calculating the cpmg trains
        intensities = {0.0: spectrometer.detect(start)}
        for ncyc in set(ncycs) - {0.0}:
            phases1, phases2 = self._get_phases(ncyc)
            echo1 = d_cp[ncyc] @ p9024090_1 @ d_cp[ncyc]
            echo2 = d_cp[ncyc] @ p9024090_2 @ d_cp[ncyc]
            cpmg1 = reduce(np.matmul, echo1[phases1])
            cpmg2 = reduce(np.matmul, echo2[phases2])
            intensities[ncyc] = spectrometer.detect(0.5 * (cpmg1 + cpmg2) @ start)

        # Return profile
        return np.array([intensities[ncyc] for ncyc in ncycs])

    @staticmethod
    def is_reference(metadata: ArrayFloat) -> ArrayBool:
        return metadata == 0


def register() -> None:
    creators = Creators(
        config_creator=CpmgHNDqZqConfig,
        spectrometer_creator=build_spectrometer,
        sequence_creator=CpmgHNDqZqSequence,
        dataset_creator=load_relaxation_dataset,
        filterer_creator=PlanesFilterer,
        printer_creator=CpmgPrinter,
        plotter_creator=CpmgPlotter,
    )
    factories.register(type=EXPERIMENT_NAME, creators=creators)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

import numpy as np

from chemex.configuration.data import RelaxationDataSettings
from chemex.configuration.experiment import (
    ExperimentConfig,
    RelaxationSettings,
    ToBeFitted,
)
from chemex.containers.dataset import load_relaxation_dataset
from chemex.experiments.factories import Creators, factories
from chemex.filterers import PlanesFilterer
from chemex.nmr.basis import Basis
from chemex.nmr.liouvillian import LiouvillianIS
from chemex.nmr.spectrometer import Spectrometer
from chemex.plotters.relaxation import RelaxationPlotter
from chemex.printers.data import RelaxationPrinter

if TYPE_CHECKING:
    from chemex.containers.data import Data
    from chemex.parameters.spin_system import SpinSystem
    from chemex.typing import ArrayBool, ArrayFloat


EXPERIMENT_NAME = "relaxation_nz"


class RelaxationNzSettings(RelaxationSettings):
    name: Literal["relaxation_nz"]
    observed_state: Literal["a", "b", "c", "d"] = "a"

    @property
    def detection(self) -> str:
        return f"[iz_{self.observed_state}]"


class RelaxationNzConfig(
    ExperimentConfig[RelaxationNzSettings, RelaxationDataSettings],
):
    @property
    def to_be_fitted(self) -> ToBeFitted:
        state = self.experiment.observed_state
        return ToBeFitted(
            rates=[f"r1_i_{state}"],
            model_free=[f"tauc_{state}", f"s2_{state}"],
        )


def build_spectrometer(
    config: RelaxationNzConfig,
    spin_system: SpinSystem,
) -> Spectrometer:
    settings = config.experiment
    conditions = config.conditions

    basis = Basis(type="iz", spin_system="nh")
    liouvillian = LiouvillianIS(spin_system, basis, conditions)
    spectrometer = Spectrometer(liouvillian)

    spectrometer.detection = settings.detection

    return spectrometer


@dataclass
class RelaxationNzSequence:
    settings: RelaxationNzSettings

    def calculate(self, spectrometer: Spectrometer, data: Data) -> ArrayFloat:
        times = data.metadata

        # Getting the starting magnetization
        start = spectrometer.get_equilibrium()

        # Return profile
        delays = spectrometer.delays(times)
        return np.array([spectrometer.detect(delay @ start) for delay in delays])

    def is_reference(self, metadata: ArrayFloat) -> ArrayBool:
        return np.full_like(metadata, False, dtype=np.bool_)


def register() -> None:
    creators = Creators(
        config_creator=RelaxationNzConfig,
        spectrometer_creator=build_spectrometer,
        sequence_creator=RelaxationNzSequence,
        dataset_creator=load_relaxation_dataset,
        filterer_creator=PlanesFilterer,
        printer_creator=RelaxationPrinter,
        plotter_creator=RelaxationPlotter,
    )
    factories.register(type=EXPERIMENT_NAME, creators=creators)



"""Module for command-line interface functionalities in ChemEx software package.

This module is an integral part of the ChemEx software, dedicated to enhancing user
interaction with a command-line interface for analyzing NMR chemical exchange data.
It leverages the 'rich' Python library to provide rich text and table formatting,
making the data presentation more readable and engaging. The module includes various
functions for displaying progress, handling errors, and showing results of data
analysis steps such as dataset loading, fitting processes, simulations, and plotting.

Typical usage example:

  from your_module import print_logo, print_loading_experiments
  print_logo()
  print_loading_experiments()
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from rich import box
from rich.console import Console
from rich.padding import Padding
from rich.panel import Panel
from rich.rule import Rule
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text

from chemex import __version__

if TYPE_CHECKING:
    from collections import Counter
    from pathlib import Path

    from pydantic import ValidationError

console = Console()


LOGO = r"""
    ________                   ______
  / ____/ /_  ___  ____ ___  / ____/  __
 / /   / __ \/ _ \/ __ `__ \/ __/ | |/_/
/ /___/ / / /  __/ / / / / / /____>  <
\____/_/ /_/\___/_/ /_/ /_/_____/_/|_|


"""
"""ASCII art logo of the ChemEx software."""

EXPERIMENT_NAME = """
[experiment]
name = "experiment_name"
...
"""
"""Template for the experiment name in configuration files."""


def print_logo() -> None:
    """Print the ChemEx software logo with version information."""
    logo = Text(LOGO, style="blue")
    description = "Analysis of NMR chemical exchange data\n\n"
    version = "Version: "
    version_number = Text(f"{__version__}", style="red")
    all_text = Text.assemble(logo, description, version, version_number)
    panel = Panel.fit(all_text)
    console.print(panel)


def print_loading_experiments() -> None:
    """Display a loading message for datasets."""
    console.print("\nLoading datasets...", style="bold yellow")


def get_reading_exp_text(filename: Path, name: str = "", total_nb: int = 0) -> Text:
    """Generate a formatted message for reading an experiment file.

    Args:
        filename (Path): The path of the file being read.
        name (str, optional): The experiment name. Defaults to an empty string.
        total_nb (int, optional): The total number of profiles. Defaults to 0.

    Returns:
        Text: A formatted message indicating the file name, type, and profile count.
    """
    parts = (
        "  • Reading ",
        (f"{filename}", "green"),
        " (",
        (f"{name}", "blue"),
        ")",
    )

    if total_nb == 0:
        return Text.assemble(*parts, " ...")

    nb_text = Text(f"{total_nb}", style="bold")

    return Text.assemble(*parts, " -> ", nb_text, " profiles")


def print_reading_defaults() -> None:
    """Print a message indicating the reading of default parameters."""
    console.print("\nReading default parameters...", style="bold yellow")


def print_reading_methods() -> None:
    """Display a message indicating that methods are being read."""
    console.print("\nReading methods...", style="bold yellow")


def print_start_fit() -> None:
    """Display a message indicating the start of the fitting process."""
    console.print("\nStarting the fits...", style="bold yellow")


def print_running_simulations() -> None:
    """Display a message indicating that simulations are running."""
    console.print("\nRunning simulations...", style="bold yellow")


def print_step_name(name: str, index: int, total: int) -> None:
    """Print the name of the current step in the analysis process.

    Args:
        name (str): Name of the step.
        index (int): Current step index.
        total (int): Total number of steps.
    """
    text = Text.assemble(
        "Running ",
        (f"{name}", "magenta"),
        f" ({index}/{total})\n",
    )
    console.print(Padding(Rule(text, end="\n\n"), (1, 0, 0, 1)), width=51)


def print_selecting_profiles(selected_nb: int) -> None:
    """Display a message indicating the number of profiles selected.

    Args:
        selected_nb (int): The number of selected profiles.
    """
    console.print(
        Text.from_markup(f"  • Selecting profiles -> [blue]{selected_nb}[/] profiles"),
    )


def print_no_data() -> None:
    """Inform the user that no data is available for fitting."""
    console.print("  • No data to fit")


def print_fitmethod(fit_method: str) -> None:
    """Display the chosen method for fitting.

    Args:
        fit_method (str): The method being used for fitting.
    """
    console.print(Text.assemble("  • Fit method -> ", (f"{fit_method}", "blue")))


def print_status_changes(
    vary: Counter[str],
    fixed: Counter[str],
    constrained: Counter[str],
) -> None:
    """Print changes in parameter status like varied, fixed, or constrained.

    Args:
        vary (Counter[str]): Parameters varied during fitting.
        fixed (Counter[str]): Parameters fixed during fitting.
        constrained (Counter[str]): Parameters constrained during fitting.
    """
    if not any([vary, fixed, constrained]):
        return

    console.print(Text("  • Setting parameter status..."))

    table = Table()
    table.add_column("Name")
    table.add_column("Status")
    table.add_column("Updated", justify="right")

    for name, nb in vary.items():
        table.add_row(name, "fitted", str(nb))

    for name, nb in fixed.items():
        table.add_row(name, "fixed", str(nb))

    for name, nb in constrained.items():
        table.add_row(name, "constrained", str(nb))

    table.box = box.SIMPLE_HEAD

    console.print(Padding.indent(table, 3))


def print_minimizing() -> None:
    """Display a message indicating that the minimization process is running."""
    console.print(Text("  • Running the minimization..."))


def print_running_grid() -> None:
    """Inform the user that the grid search is in progress."""
    console.print(Text("  • Running the grid search..."))


def print_running_statistics(name: str) -> None:
    """Display a message indicating that statistical simulations are running.

    Args:
        name (str): The name of the statistical simulation.
    """
    console.print(Text(f"  • Running {name} simulations..."))


def print_section(name: str) -> None:
    """Print the name of the current section in the analysis process.

    Args:
        name (str): Name of the section.
    """
    console.print(Text(f"  • Section {name}"))


def print_chi2_table_header() -> None:
    """Print the header for the chi-squared table."""
    header = Text(f"{'Iteration':>9s}  {'χ²':>12s}  {'Reduced χ²':>12s}")
    console.print()
    console.print(Padding.indent(header, 5), style="bold")
    console.print(Padding.indent("─" * 39, 4))


def print_chi2_table_line(iteration: int, chisqr: float, redchi: float) -> None:
    """Print a line in the chi-squared table with iteration and chi-squared values.

    Args:
        iteration (int): Iteration number.
        chisqr (float): Chi-squared value.
        redchi (float): Reduced chi-squared value.
    """
    line = Text(f"{iteration:>9d}  {chisqr:>12.1f}  {redchi:>12.3f}")
    console.print(Padding.indent(line, 5))


def print_chi2_table_footer(iteration: int, chisqr: float, redchi: float) -> None:
    """Print the footer for the chi-squared table.

    Args:
        iteration (int): Iteration number.
        chisqr (float): Chi-squared value.
        redchi (float): Reduced chi-squared value.
    """
    footer = Text(f"{iteration:>9d}  {chisqr:>12.1f}  {redchi:>12.3f}")
    console.print(Padding.indent("─" * 39, 4))
    console.print(Padding.indent(footer, 5), style="bold")
    console.print()


def print_chi2(chisqr: float, redchi: float) -> None:
    """Display the chi-squared and reduced chi-squared values.

    Args:
        chisqr (float): Chi-squared value.
        redchi (float): Reduced chi-squared value.
    """
    console.print()
    console.print(
        Padding.indent(Text.from_markup(f"        χ²: [bold]{chisqr:15.1f}[/]"), 3),
    )
    console.print(
        Padding.indent(Text.from_markup(f"Reduced χ²: [bold]{redchi:15.3f}[/]"), 3),
    )
    console.print()


def print_writing_results(path: Path) -> None:
    """Inform the user about the location where results are being written.

    Args:
        path (Path): Path to the file or directory where results are saved.
    """
    console.print(f"  • Writing results in [green]{path}")


def print_making_plots() -> None:
    """Display a message indicating that plots are being generated."""
    console.print(Text("  • Making plots..."))


def print_plot_filename(filename: Path, *, extra: bool = True) -> None:
    """Display the filename of the plot being generated.

    Args:
        filename (Path): The path of the plot file.
        extra (bool, optional): Indicates if extra information is included.
                                Defaults to True.
    """
    text = f"    ‣ [green]{filename}[/]"

    if extra:
        text += " [[green].fit[/], [green].exp[/]]"

    console.print(Text.from_markup(text))


def print_group_name(text: str) -> None:
    """Print the name of the group in the analysis process.

    Args:
        text (str): Name of the group.
    """
    console.print(Padding(Rule(text, characters="⋅"), (1, 0, 0, 3)), width=49)


def print_file_not_found(filename: Path) -> None:
    """Inform the user that a specified file was not found.

    Args:
        filename (Path): The path of the file that was not found.
    """
    console.print()
    console.print(f"[red]The file '{filename}' is empty or does not exist!")


def print_file_not_found_error(error: FileNotFoundError) -> None:
    """Display the file not found error message.

    Args:
        error (FileNotFoundError): The FileNotFoundError exception instance.
    """
    console.print()
    console.print(f"[red]Error: {error}")


def print_toml_error(filename: Path, error_message: Exception) -> None:
    """Display an error related to TOML file parsing.

    Args:
        filename (Path): Path of the TOML file.
        error_message (Exception): The exception instance with the error message.
    """
    console.print()
    console.print(Text.from_markup(f"[red]Error in the TOML file '{filename}'"))
    console.print(
        Text.from_markup(f"[red]-> {error_message}\n"),
    )
    console.print(
        "Please check the website [link]https://toml.io[/]"
        " for a complete description of the TOML format.",
    )


def print_experiment_name_error(filename: Path) -> None:
    """Display an error message when the experiment name is missing from a file.

    Args:
        filename (Path): Path of the file with the missing experiment name.
    """
    console.print()
    console.print(f"[red]The experiment name is missing from the file '{filename}'\n")
    console.print(
        f"Please make sure that the 'name' field is provided in '{filename}':\n",
    )
    console.print(
        Syntax(EXPERIMENT_NAME, "toml"),
        "\nRun 'chemex info' to obtain the full list of the available experiments.",
    )


def print_pydantic_parsing_error(filename: Path, error: ValidationError) -> None:
    """Display errors encountered while parsing a file using Pydantic.

    Args:
        filename (Path): Path of the file being parsed.
        error (ValidationError): Instance detailing parsing errors.
    """
    console.print()
    console.print(Text.from_markup(f"[red]Error(s) while parsing '{filename}'"))
    for err in error.errors():
        location = " -> ".join(str(loc) for loc in err["loc"])
        console.print("  • ", location, ":", err["msg"], style="red")


def print_calculation_stopped_error() -> None:
    """Inform the user that the calculation was manually stopped."""
    console.print()
    console.print("[red] -- Keyboard Interrupt: Calculation stopped --")
    console.print()


def print_plotting_canceled() -> None:
    """Display a message indicating that the plotting process was cancelled."""
    console.print()
    console.print("[red] -- Keyboard Interrupt: Plotting cancelled --")
    console.print()


def print_value_error() -> None:
    """Inform the user that a ValueError occurred, stopping the calculation."""
    console.print()
    console.print("[red] -- Got a ValueError: Calculation stopped --")
    console.print()


def print_warning_positive_jnh() -> None:
    """Warn about positive 1J(NH) coupling values."""
    console.print()
    console.print(
        "[yellow] -- WARNING: Some 1J(NH) couplings are set with positive values --",
    )
    console.print(
        "This can cause the TROSY and anti-TROSY components to be switched in some"
        " experiments.",
    )
    console.print()


def print_warning_negative_jch() -> None:
    """Warn about negative 1J(CH) coupling values."""
    console.print()
    console.print(
        "[yellow] -- WARNING: Some 1J(CH) couplings are set with negative values --",
    )
    console.print(
        "This can cause the TROSY and anti-TROSY components to be switched in some"
        " experiments.",
    )
    console.print()


def print_error_grid_settings(entry: str) -> None:
    """Display an error related to grid settings.

    Args:
        entry (str): The problematic entry in the grid settings.
    """
    console.print()
    console.print("[red] -- ERROR: Error reading grid settings:")
    console.print(Text(f'    "{entry}"', style="red"))
    console.print()
    console.print(
        "Please make sure that the grid settings are provided in the correct format",
    )
    console.print(Text("    [PB] = lin(<min>, <max>, <nb of points>)"))
    console.print(Text("    [PB] = log(<min>, <max>, <nb of points>)"))
    console.print(Text("    [PB] = (<value1>, <value2>, ..., <valuen>)"))
    console.print()


def print_error_constraints(expression: str) -> None:
    """Display an error message for issues with constraints expressions.

    Args:
        expression (str): The problematic constraint expression.
    """
    console.print()
    console.print(f'[red] -- ERROR: Error reading constraints -> "{expression}" --')
    console.print()


def print_grid_statistic_warning() -> None:
    """Warn that 'GRID' and 'STATISTICS' options are mutually exclusive."""
    console.print()
    console.print(
        "[yellow] -- WARNING: The 'GRID' and 'STATISTICS' options are mutually "
        "exclusive. Only the 'GRID' calculation will be run.",
    )
    console.print()


def print_model_error(name: str) -> None:
    """Display an error message for unavailable models.

    Args:
        name (str): The name of the model that is not available.
    """
    from chemex.models.factory import model_factory

    console.print()
    console.print(f"[red] -- ERROR: The model '{name}' is not available.")
    console.print()
    console.print("The available models are:")
    for model_name in sorted(model_factory.set):
        console.print(f"    - '{model_name}'")
    console.print()


def print_not_implemented_noise_method_warning(
    filename: Path,
    kind: str,
    implemented: tuple[str, ...],
) -> None:
    """Warn about unimplemented noise methods for an experiment.

    Args:
        filename (Path): Path of the experiment file.
        kind (str): The kind of noise method that is not implemented.
        implemented (tuple[str, ...]): Tuple of implemented methods.
    """
    warning_message = (
        f"[yellow] -- WARNING: Experiment {filename.name}[/yellow]: "
        f"The '{kind}' method is not implemented. "
        f"Please choose one of the following methods: {', '.join(implemented)}. "
        f"Defaulting to 'file' method."
    )
    console.print()
    console.print(warning_message)
    console.print()


def print_no_duplicate_warning(filename: Path) -> None:
    """Warn about the absence of duplicate points in some profiles.

    Args:
        filename (Path): Path of the experiment file.
    """
    warning_message = (
        f"[yellow] -- WARNING: Experiment {filename.name}[/yellow]: "
        f"Some profiles do not have duplicate points. "
        f"Uncertainties cannot be estimated and will be directly taken from the files. "
        f"Defaulting to 'file' method."
    )
    console.print()
    console.print(warning_message)
    console.print()


FITMETHOD_ERROR_MESSAGE = """\
  - "FITMETHOD" must be one of:
        "leastsq": Levenberg-Marquardt (default)
        "least_squares": least-squares minimization (Trust Region Reflective method)
        "differential_evolution": differential evolution
        "brute": brute force method
        "basinhopping": basinhopping
        "ampgo": Adaptive Memory Programming for Global Optimization
        "nelder": Nelder-Mead
        "lbfgsb": L-BFGS-B
        "powell": Powell
        "cg": Conjugate-Gradient
        "newton": Newton-CG
        "cobyla": Cobyla
        "bfgs": BFGS
        "tnc": Truncated Newton
        "trust-ncg": Newton-CG trust-region
        "trust-exact": nearly exact trust-region
        "trust-krylov": Newton GLTR trust-region
        "trust-constr": trust-region for constrained optimization
        "dogleg": Dog-leg trust-region
        "slsqp": Sequential Linear Squares Programming
        "emcee": Maximum likelihood via Monte-Carlo Markov Chain
        "shgo": Simplicial Homology Global Optimization
        "dual_annealing": Dual Annealing optimization"""

INCLUDE_ERROR_MESSAGE = """\
  - "INCLUDE" must be a list of nuclei to be included or '*' to select all the profiles.
    Example: ['G3H', 'S4H', 'W5H'] or [3, 4, 5]"""

EXCLUDE_ERROR_MESSAGE = """\
  - "EXCLUDE" must be a list of nuclei to be excluded.

    Example: ['G3H', 'S4H', 'W5H'] or [3, 4, 5]"""

FIT_ERROR_MESSAGE = """\
  - "FIT" must be a list of parameters to be fitted.

    Example: ["PB", "KEX_AB"]"""

FIX_ERROR_MESSAGE = """\
  - "FIX" must be a list of parameters to be fixed.

    Example: ["PB", "KEX_AB"]"""

CONSTRAINTS_ERROR_MESSAGE = """\
  - "CONSTRAINTS" must be a list of expressions.

    Example: ["[R2_B] = [R2_A] / 2", "[R1_B] = [R1_A]"]"""

GRID_ERROR_MESSAGE = """\
  - "GRID" must be a list of parameters to be gridded with the associated grid
    definition.

    Example: GRID    = [
        "[KEX_AB] = log(100.0, 600.0, 10)",
        "[PB] = log(0.03, 0.15, 10)",
        "[DW_AB] = lin(0.0, 10.0, 5)",
    ]"""

STATISTICS_ERROR_MESSAGE = """\
  - "STATISTICS" must be a dictionary with keys 'MC', 'BS', 'BSN'

    Example: { "MC"=10 } or { "MC"=10, "BS"=10 }"""

METHOD_ERROR_MESSAGES = {
    "fitmethod": FITMETHOD_ERROR_MESSAGE,
    "include": INCLUDE_ERROR_MESSAGE,
    "exclude": EXCLUDE_ERROR_MESSAGE,
    "fit": FIT_ERROR_MESSAGE,
    "fix": FIX_ERROR_MESSAGE,
    "constraints": CONSTRAINTS_ERROR_MESSAGE,
    "grid": GRID_ERROR_MESSAGE,
    "statistics": STATISTICS_ERROR_MESSAGE,
}


def print_wrong_option(option: str) -> str:
    """Display a message for an invalid option in configuration files.

    Args:
        option (str): The option that is invalid or incorrectly formatted.

    Returns:
        str: A message indicating the error with the specific option.
    """
    return (
        f"\n  - '{option.upper()}' is not a valid option.\n"
        "\nPlease check the website "
        "[link]"
        "https://gbouvignies.github.io/ChemEx/docs/user_guide/fitting/method_files"
        "[/] "
        "for a complete list of valid options."
    )


def print_method_error(filename: Path, section: str, options: set[int | str]) -> None:
    """Display errors for invalid methods or options in a specific section of a file.

    Args:
        filename (Path): The file containing the error.
        section (str): The section of the file with the erroneous method or option.
        options (set[int | str]): Set of options or methods that are incorrect.
    """
    console.print()
    console.print(
        f"[red] -- ERROR: Error reading section '[{section}]' of '{filename}' --",
    )
    for option in options:
        console.print(
            METHOD_ERROR_MESSAGES.get(str(option), print_wrong_option(str(option))),
        )
    console.print()



from __future__ import annotations

from collections import defaultdict
from contextlib import ExitStack
from itertools import product
from typing import TYPE_CHECKING, Any

import numpy as np
from matplotlib.backends.backend_pdf import PdfPages

from chemex.containers.data import Data
from chemex.messages import print_plot_filename
from chemex.plotters.plot import get_grid, plot_profile
from chemex.printers.plot import PlotPrinter, data_plot_printers

if TYPE_CHECKING:
    from pathlib import Path

    from chemex.containers.profile import Profile


def plot_exsy(file_pdf: PdfPages, name: str, data_exp: Data, data_calc: Data):
    fig = plot_profile(name, data_exp, data_calc)
    ax2 = fig.axes[1]
    ax2.set_xlabel(r"Time (s)")
    ax2.set_ylabel(r"Intensity")
    ax2.set_xlim(0.0, max(data_exp.metadata) + min(data_calc.metadata))
    file_pdf.savefig(fig)


def create_plot_data_exp(profile: Profile) -> dict[tuple[str, str], Data]:
    data = profile.data

    intensity0 = data.exp[np.argmax(abs(data.exp))]

    data_values = defaultdict(list)
    for metadata, exp, calc, err, mask in zip(
        data.metadata,
        data.exp,
        data.calc,
        data.err,
        data.mask,
        strict=True,
    ):
        states = tuple(metadata[["states1", "states2"]])
        data_values[states].append((metadata["times"], exp, calc, err, mask))

    data_exp_dict = {}
    dtype = [
        ("times", "f8"),
        ("exp", "f8"),
        ("calc", "f8"),
        ("err", "f8"),
        ("mask", "?"),
    ]
    for states, values in data_values.items():
        values_array = np.array(values, dtype=dtype)
        values_array["exp"] /= intensity0
        values_array["calc"] /= intensity0
        errors = values_array["err"] / abs(intensity0)
        errorbars = np.array([-errors, errors]).transpose()

        data_exp = Data(
            exp=values_array["exp"],
            err=errorbars,
            metadata=values_array["times"],
        )
        data_exp.calc = values_array["calc"]
        data_exp.mask = values_array["mask"]
        data_exp.sort()

        data_exp_dict[states] = data_exp

    return data_exp_dict


def create_plot_data_calc(profile: Profile) -> dict[tuple[str, str], Data]:
    spectrometer = profile.spectrometer
    data = profile.data

    times = get_grid(data.metadata["times"], 100, 0.0)
    states = np.unique(data.metadata[["states1", "states2"]])
    metadata_list = [
        (time, state1, state2) for time, (state1, state2) in product(times, states)
    ]
    dtype = [("times", "f8"), ("states1", "U1"), ("states2", "U1")]
    metadata = np.array(metadata_list, dtype=dtype)

    filler = np.zeros_like(metadata["times"])
    data_fit = Data(exp=filler, err=filler, metadata=metadata)

    intensity0 = data.exp[np.argmax(abs(data.exp))]
    scale = profile.data.scale / intensity0
    data_fit.calc = scale * profile.pulse_sequence.calculate(spectrometer, data_fit)

    data_values = defaultdict(list)
    for metadata, calc in zip(data_fit.metadata, data_fit.calc):
        states = tuple(metadata[["states1", "states2"]])
        data_values[states].append((metadata["times"], calc))

    data_calc_dict = {}
    dtype = [("times", "f8"), ("calc", "f8")]
    for states, values in data_values.items():
        values_array = np.array(values, dtype=dtype)
        filler = np.zeros_like(values_array["calc"])

        data_calc = Data(exp=filler, err=filler, metadata=values_array["times"])
        data_calc.calc = values_array["calc"]
        data_calc.sort()

        data_calc_dict[states] = data_calc

    return data_calc_dict


class EXSYPlotter:
    def __init__(self, filename: Path, **_extra: Any) -> None:
        self.filename = filename
        self.printer: PlotPrinter = data_plot_printers["relaxation"]

    def plot(self, path: Path, profiles: list[Profile]) -> None:
        basename = path / self.filename.name
        name_pdf = basename.with_suffix(".pdf")
        name_exp = basename.with_suffix(".exp")
        name_fit = basename.with_suffix(".fit")

        print_plot_filename(name_pdf)

        with ExitStack() as stack:
            file_pdf = stack.enter_context(PdfPages(str(name_pdf)))
            file_calc = stack.enter_context(name_fit.open("w"))
            file_exp = stack.enter_context(name_exp.open("w"))
            for profile in sorted(profiles):
                data_exp_dict = create_plot_data_exp(profile)
                data_calc_dict = create_plot_data_calc(profile)
                for states in data_exp_dict:
                    data_exp = data_exp_dict[states]
                    data_calc = data_calc_dict[states]
                    name = f"{profile.spin_system!s}_{''.join(states)}"
                    plot_exsy(file_pdf, name, data_exp, data_calc)
                    file_exp.write(self.printer.print_exp(name, data_exp))
                    file_calc.write(self.printer.print_calc(name, data_calc))



from __future__ import annotations

from dataclasses import dataclass
from functools import reduce
from typing import TYPE_CHECKING, Literal

import numpy as np

from chemex.configuration.data import RelaxationDataSettings
from chemex.configuration.experiment import CpmgSettings, ExperimentConfig, ToBeFitted
from chemex.containers.dataset import load_relaxation_dataset
from chemex.experiments.factories import Creators, factories
from chemex.filterers import PlanesFilterer
from chemex.nmr.basis import Basis
from chemex.nmr.constants import GAMMA
from chemex.nmr.liouvillian import LiouvillianIS
from chemex.nmr.spectrometer import Spectrometer
from chemex.plotters.cpmg import CpmgPlotter
from chemex.printers.data import CpmgPrinter

if TYPE_CHECKING:
    from chemex.containers.data import Data
    from chemex.parameters.spin_system import SpinSystem
    from chemex.typing import ArrayBool, ArrayFloat, ArrayInt


EXPERIMENT_NAME = "cpmg_ch3_1h_tq_diff"


class CpmgCh31HTqDiffSettings(CpmgSettings):
    name: Literal["cpmg_ch3_1h_tq_diff"]
    time_t2: float
    carrier: float
    pw90: float
    delta: float
    gradient: float
    tau: float = 500.0e-6
    tauc: float = 0.67e-3  # ~ 1/(12*J[HC])
    comp180_flg: bool = True
    ipap_flg: bool = False
    observed_state: Literal["a", "b", "c", "d"] = "a"

    @property
    def k2_factor(self) -> float:
        return (3.0 * GAMMA["h"] * self.gradient * self.delta) ** 2

    @property
    def detection(self) -> str:
        return f"[2ixsz_{self.observed_state}]"


class CpmgCh31HTqDiffConfig(
    ExperimentConfig[CpmgCh31HTqDiffSettings, RelaxationDataSettings],
):
    @property
    def to_be_fitted(self) -> ToBeFitted:
        state = self.experiment.observed_state

        to_be_fitted = ToBeFitted(
            rates=[f"r2_i_{state}", "d_"],
            model_free=[f"tauc_{state}", "d_"],
        )

        if self.experiment.ipap_flg:
            to_be_fitted.rates.append(f"r2a_i_{state}")
            to_be_fitted.model_free.append(f"s2_{state}")

        return to_be_fitted


def build_spectrometer(
    config: CpmgCh31HTqDiffConfig,
    spin_system: SpinSystem,
) -> Spectrometer:
    settings = config.experiment
    conditions = config.conditions

    basis = Basis(type="ixyzsz_diff", extension="tq", spin_system="hc")
    liouvillian = LiouvillianIS(spin_system, basis, conditions)
    spectrometer = Spectrometer(liouvillian)

    spectrometer.carrier_i = settings.carrier
    spectrometer.b1_i = 1 / (4.0 * settings.pw90)
    spectrometer.detection = settings.detection

    return spectrometer


@dataclass
class CpmgCh31HTqDiffSequence:
    settings: CpmgCh31HTqDiffSettings

    def _get_delays(self, ncycs: ArrayFloat) -> tuple[dict[float, float], list[float]]:
        ncyc_no_ref = ncycs[ncycs > 0]
        factor = 2.0 if self.settings.comp180_flg else 1.0
        tau_cps = {
            ncyc: self.settings.time_t2 / (4.0 * ncyc) - factor * self.settings.pw90
            for ncyc in ncyc_no_ref
        }
        delays = [self.settings.tau, *tau_cps.values()]
        return tau_cps, delays

    def _get_phases(self, ncyc: float) -> tuple[ArrayInt, ArrayInt]:
        cp_phases1 = [0, 1, 0, 1, 1, 0, 1, 0, 2, 3, 2, 3, 3, 2, 3, 2]
        cp_phases2 = [0, 3, 0, 3, 3, 0, 3, 0, 2, 1, 2, 1, 1, 2, 1, 2]
        indexes = np.arange(int(ncyc))
        phases1 = np.take(cp_phases1, np.flip(indexes), mode="wrap")
        phases2 = np.take(cp_phases2, indexes, mode="wrap")
        return phases1, phases2

    def calculate(self, spectrometer: Spectrometer, data: Data) -> ArrayFloat:
        ncycs = data.metadata

        # Getting the starting magnetization
        start = spectrometer.get_start_magnetization(["2ixsz"], atom="h")

        # Calculation of the spectrometers with gradient 0
        spectrometer.gradient_dephasing = 0.0
        d_tau_0, d_tauc_0 = spectrometer.delays([self.settings.tau, self.settings.tauc])
        p180_sx = spectrometer.perfect180_s[0]

        # Calculation of the spectrometers with gradient 1
        spectrometer.gradient_dephasing = 1.0 / 3.0 * self.settings.k2_factor
        (d_delta_1,) = spectrometer.delays(self.settings.delta)

        # Calculation of the spectrometers with gradient 2
        spectrometer.gradient_dephasing = self.settings.k2_factor
        (d_tau_2,) = spectrometer.delays(self.settings.tau)
        p90_2 = spectrometer.p90_i
        p180_2 = spectrometer.p180_i

        # Calculation of the spectrometers with gradient 3
        spectrometer.gradient_dephasing = 7.0 / 3.0 * self.settings.k2_factor
        (d_delta_3,) = spectrometer.delays(self.settings.delta)

        # Calculation of the spectrometers with gradient 4
        spectrometer.gradient_dephasing = 4.0 * self.settings.k2_factor
        tau_cps, all_delays = self._get_delays(ncycs)
        delays_4 = dict(zip(all_delays, spectrometer.delays(all_delays), strict=True))
        d_cp_4 = {ncyc: delays_4[delay] for ncyc, delay in tau_cps.items()}
        d_tau_4 = delays_4[self.settings.tau]
        p180_4 = spectrometer.p180_i
        p180pmy_4 = 0.5 * (p180_4[1] + p180_4[3])  # +/- phase cycling

        if self.settings.comp180_flg:
            p180_cp1 = spectrometer.p9018090_i_1
            p180_cp2 = spectrometer.p9018090_i_2
        else:
            p180_cp1 = p180_cp2 = p180_4

        # The liouvillian for evolution before and after the CPMG period
        p180_grad1, p180_grad2 = p90_2[[1, 3]] @ p180_2[0] @ p90_2[[1, 3]]
        grad1 = d_tau_4 @ d_delta_3 @ p180_grad1 @ d_tau_2 @ d_delta_1
        grad2 = d_tau_0 @ d_delta_1 @ p180_grad2 @ d_tau_2 @ d_delta_3
        if self.settings.ipap_flg:
            part1 = d_tauc_0 @ start
            part2 = d_tauc_0
            centre0 = grad2 @ p180pmy_4 @ grad1
            centre0 = 0.5 * (centre0 + p180_sx @ centre0 @ p180_sx)
        else:
            part1 = start
            part2 = spectrometer.identity
            centre0 = grad2 @ p180pmy_4 @ grad1

        # Calculating the intensities as a function of ncyc
        intst = {0.0: spectrometer.detect(part2 @ centre0 @ part1)}

        for ncyc in set(ncycs) - {0.0}:
            phases1, phases2 = self._get_phases(ncyc)
            echo1 = d_cp_4[ncyc] @ p180_cp1 @ d_cp_4[ncyc]
            echo2 = d_cp_4[ncyc] @ p180_cp2 @ d_cp_4[ncyc]
            cpmg1 = reduce(np.matmul, echo1[phases1])
            cpmg2 = reduce(np.matmul, echo2[phases2])
            centre = grad2 @ cpmg2 @ p180pmy_4 @ cpmg1 @ grad1
            if self.settings.ipap_flg:
                centre = 0.5 * (centre + p180_sx @ centre @ p180_sx)
            intst[ncyc] = spectrometer.detect(part2 @ centre @ part1)

        # Return profile
        return np.array([intst[ncyc] for ncyc in ncycs])

    @staticmethod
    def is_reference(metadata: ArrayFloat) -> ArrayBool:
        return metadata == 0


def register() -> None:
    creators = Creators(
        config_creator=CpmgCh31HTqDiffConfig,
        spectrometer_creator=build_spectrometer,
        sequence_creator=CpmgCh31HTqDiffSequence,
        dataset_creator=load_relaxation_dataset,
        filterer_creator=PlanesFilterer,
        printer_creator=CpmgPrinter,
        plotter_creator=CpmgPlotter,
    )
    factories.register(type=EXPERIMENT_NAME, creators=creators)



from __future__ import annotations

from dataclasses import dataclass, field
from itertools import chain
from typing import TYPE_CHECKING, TypeVar

import numpy as np

from chemex.messages import (
    print_no_duplicate_warning,
    print_not_implemented_noise_method_warning,
)
from chemex.parameters import database
from chemex.uncertainty import estimate_noise_variance

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from lmfit import Parameters as ParametersLF

    from chemex.configuration.methods import Selection
    from chemex.containers.profile import Profile
    from chemex.parameters.spin_system import Group
    from chemex.plotters.plotter import Plotter
    from chemex.printers.data import Printer

Self = TypeVar("Self", bound="Experiment")


@dataclass
class Experiment:
    filename: Path
    name: str
    profiles: list[Profile]
    filtered_profiles: list[Profile] = field(init=False, default_factory=list)
    printer: Printer
    plotter: Plotter

    def residuals(self, params: ParametersLF) -> list[float]:
        return list(
            chain.from_iterable(profile.residuals(params) for profile in self.profiles),
        )

    def plot(self, path: Path):
        self.plotter.plot(path, self.profiles)

    def plot_simulation(self, path: Path):
        self.plotter.plot_simulation(path, self.profiles)

    def write(self, path: Path):
        filename = (path / self.filename.name).with_suffix(".dat")
        with filename.open("w") as file_dat:
            file_dat.write(self.printer.header)
            for profile in sorted(self.profiles):
                file_dat.write(str(profile))

    def select(self, selection: Selection):
        include = selection.include
        exclude = selection.exclude
        profiles_all = [*self.profiles, *self.filtered_profiles]
        profiles: list[Profile] = []
        filtered: list[Profile] = []
        for profile in profiles_all:
            included = include is None or profile.spin_system.part_of(include)
            excluded = exclude is not None and profile.spin_system.part_of(exclude)
            if included and not excluded:
                profiles.append(profile)
            else:
                filtered.append(profile)
        self.profiles = profiles
        self.filtered_profiles = filtered

    def filter(self, params: ParametersLF):
        for profile in self.profiles:
            profile.filter(params)

    def _any_duplicate(self):
        return any(profile.any_duplicate() for profile in self.profiles)

    def estimate_noise(self, kind: str) -> None:
        # TODO: Validation should be moved to the configuration file module
        implemented = ("file", "scatter", "duplicates")
        if kind not in implemented:
            print_not_implemented_noise_method_warning(self.filename, kind, implemented)
            kind = "file"
        if kind == "duplicates" and not self._any_duplicate():
            print_no_duplicate_warning(self.filename)
            kind = "file"
        if kind == "file" or not self.profiles:
            return
        noise_variance_values = [
            estimate_noise_variance[kind](profile.data) for profile in self.profiles
        ]
        noise_mean = np.sqrt(np.mean(noise_variance_values))
        for profile in self.profiles:
            profile.set_noise(noise_mean)

    def prepare_for_simulation(self) -> None:
        for profile in self.profiles:
            profile.prepare_for_simulation()

    def monte_carlo(self: Self) -> Self:
        profiles = [profile.monte_carlo() for profile in self.profiles]
        return type(self)(
            self.filename,
            self.name,
            profiles,
            self.printer,
            self.plotter,
        )

    def bootstrap(self: Self) -> Self:
        profiles = [profile.bootstrap() for profile in self.profiles]
        return type(self)(
            self.filename,
            self.name,
            profiles,
            self.printer,
            self.plotter,
        )

    def bootstrap_ns(self: Self, groups: list[Group]) -> Self:
        """Residue-specific bootstrap."""
        profiles: dict[Group, list[Profile]] = {}
        for profile in self.profiles:
            profiles.setdefault(profile.spin_system.groups["i"], []).append(profile)
        profiles_bs_ns: list[Profile] = []
        for group in groups:
            profiles_bs_ns.extend(profiles.get(group, []))
        return type(self)(
            self.filename,
            self.name,
            profiles_bs_ns,
            self.printer,
            self.plotter,
        )

    @property
    def groups(self) -> set[Group]:
        return {profile.spin_system.groups["i"] for profile in self.profiles}

    @property
    def param_id_sets(self) -> list[set[str]]:
        return [
            set(database.get_parameters(profile.param_ids)) for profile in self.profiles
        ]

    def get_relevant_subset(self: Self, param_ids: set[str]) -> Self:
        profiles = [
            profile
            for profile in self.profiles
            if set(database.get_parameters(profile.param_ids)) & param_ids
        ]

        return type(self)(
            self.filename,
            self.name,
            profiles,
            self.printer,
            self.plotter,
        )

    def __iter__(self) -> Iterator[Profile]:
        yield from self.profiles

    def __len__(self) -> int:
        return len(self.profiles)

    def __bool__(self) -> bool:
        return bool(len(self.profiles))



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field
from functools import cached_property
from operator import attrgetter
from typing import TYPE_CHECKING, Protocol, TypeVar

from cachetools import LRUCache, cachedmethod

if TYPE_CHECKING:
    from collections.abc import Hashable

    from lmfit import Parameters as ParametersLF

    from chemex.containers.data import Data
    from chemex.nmr.spectrometer import Spectrometer
    from chemex.parameters.spin_system import SpinSystem
    from chemex.printers.data import Printer
    from chemex.typing import ArrayBool, ArrayFloat

Self = TypeVar("Self", bound="Profile")


class PulseSequence(Protocol):
    """Defines a protocol for pulse sequence classes."""

    def calculate(self, spectrometer: Spectrometer, data: Data) -> ArrayFloat:
        """Calculate the ArrayFloat given a Spectrometer and Data."""
        ...

    def is_reference(self, metadata: ArrayFloat) -> ArrayBool:
        """Check if the metadata is a reference."""
        ...


class Filterer(Protocol):
    """Defines a protocol for filterer classes."""

    def filter(self, data: Data) -> None:
        """Filter the given Data."""
        ...


@dataclass(order=True)
class Profile:
    """Represents the profile for a specific experiment."""

    data: Data = field(compare=False)
    spectrometer: Spectrometer = field(compare=False)
    pulse_sequence: PulseSequence = field(compare=False)
    name_map: dict[str, str] = field(compare=False)
    printer: Printer = field(compare=False)
    filterer: Filterer | None = field(compare=False, default=None)
    is_scaled: bool = field(compare=False, default=True)
    spin_system: SpinSystem = field(compare=True, init=False)
    cache: LRUCache = field(compare=False, init=False)

    def _cache_key(self, params: ParametersLF) -> tuple[Hashable, ...]:
        return (
            *(params[param_id].value for param_id in self.param_ids),
            self.data.metadata.tostring(),
        )

    def __post_init__(self):
        """Initialize derived attributes."""
        self.spin_system = self.spectrometer.liouvillian.spin_system
        self.data.refs = self.pulse_sequence.is_reference(self.data.metadata)
        self.cache = LRUCache(maxsize=5)

    @cached_property
    def param_ids(self) -> set[str]:
        """Get the set of parameter IDs."""
        return set(self.name_map.values())

    def _get_parameter_values(self, params: ParametersLF) -> dict[str, float]:
        """Get the parameter values from the provided parameters."""
        return {
            local_name: params[param_id].value
            for local_name, param_id in self.name_map.items()
        }

    def update_spectrometer(self, params: ParametersLF) -> None:
        """Update the spectrometer settings."""
        parameter_values = self._get_parameter_values(params)
        self.spectrometer.update(parameter_values)

    def calculate(self, params: ParametersLF) -> ArrayFloat:
        """Calculate and return the ArrayFloat."""
        self.update_spectrometer(params)
        self.data.calc = self.pulse_sequence.calculate(self.spectrometer, self.data)
        if self.is_scaled:
            self.data.scale_calc()
        return self.data.calc

    @cachedmethod(attrgetter("cache"), key=_cache_key)
    def residuals(self, params: ParametersLF) -> ArrayFloat:
        """Calculate and return residuals."""
        residuals = (self.calculate(params) - self.data.exp) / self.data.err
        return residuals[self.data.mask]

    def filter(self, params: ParametersLF) -> None:
        """Apply a filter to the data, if a filterer is available."""
        if self.filterer is not None:
            self.update_spectrometer(params)
            self.filterer.filter(self.data)

    def set_noise(self, value: float):
        """Set the noise value."""
        self.data.err[:] = value

    def prepare_for_simulation(self) -> None:
        """Prepare data for simulation."""
        self.data.prepare_for_simulation()

    def monte_carlo(self: Self) -> Self:
        """Generate a Monte Carlo variant of the profile."""
        profile = deepcopy(self)
        profile.data = profile.data.monte_carlo()
        return profile

    def bootstrap(self: Self) -> Self:
        """Generate a bootstrap variant of the profile."""
        profile = deepcopy(self)
        profile.data = profile.data.bootstrap()
        return profile

    def any_duplicate(self):
        """Check for duplicate data points."""
        return self.data.any_duplicate()

    def __add__(self: Self, other: Self) -> Self:
        """Combine two profiles."""
        profile = deepcopy(self)
        profile.data = self.data + other.data
        return profile

    def __str__(self) -> str:
        """String representation of the Profile."""
        return self.printer.print(str(self.spin_system), self.data)



from __future__ import annotations

"""This code imports the necessary modules and functions to define a 3-state
double binding model. The model is named "3st_double_binding" and it is
registered in the model factory. The model requires two parameters,
"p_total" and "l_total", to be specified in the conditions. The code
also defines a function, "calculate_concentrations", which calculates
the concentrations of free protein and protein bound to each ligand
given the total protein and ligand concentrations, as well as the
dissociation constants for each binding site.
"""

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np
from scipy.optimize import root

from chemex.models.constraints import pop_3st
from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions
    from chemex.typing import ArrayFloat

NAME = "3st_double_binding"

TPL = ("temperature", "p_total", "l_total")


def calculate_residuals(
    populations: ArrayFloat,
    p_total: float,
    l_total: float,
    kd_ab: float,
    kd_ac: float,
) -> ArrayFloat:
    pfree, pl1, pl2 = populations
    lfree = l_total - p_total + pfree
    return np.array(
        [
            pfree + pl1 + pl2 - p_total,
            lfree * pfree - kd_ab * pl1,
            lfree * pfree - kd_ac * pl2,
        ],
    )


@lru_cache(maxsize=100)
def calculate_concentrations(
    p_total: float,
    l_total: float,
    kd_ab: float,
    kd_ac: float,
) -> dict[str, float]:
    concentrations_start = (p_total, 0.0, 0.0)
    results = root(
        calculate_residuals,
        concentrations_start,
        args=(p_total, l_total, kd_ab, kd_ac),
    )
    return dict(zip(("pfree", "pl1", "pl2"), results["x"], strict=True))


def make_settings_3st_double_binding(
    conditions: Conditions,
) -> dict[str, ParamLocalSetting]:
    p_total = conditions.p_total
    l_total = conditions.l_total
    if p_total is None:
        msg = f"'p_total' must be specified to use the '{NAME}' model"
        raise ValueError(msg)
    if l_total is None:
        msg = f"'l_total' must be specified to use the '{NAME}' model"
        raise ValueError(msg)
    return {
        "koff_ab": ParamLocalSetting(
            name_setting=NameSetting("koff_ab", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "kd_ab": ParamLocalSetting(
            name_setting=NameSetting("kd_ab", "", ("temperature",)),
            value=1e-3,
            min=0.0,
            vary=True,
        ),
        "koff_ac": ParamLocalSetting(
            name_setting=NameSetting("koff_ac", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "kd_ac": ParamLocalSetting(
            name_setting=NameSetting("kd_ac", "", ("temperature",)),
            value=1e-3,
            min=0.0,
            vary=True,
        ),
        "kon_ab": ParamLocalSetting(
            name_setting=NameSetting("kon_ab", "", ("temperature",)),
            expr="{koff_ab} / max({kd_ab}, 1e-100)",
        ),
        "kon_ac": ParamLocalSetting(
            name_setting=NameSetting("kon_ac", "", ("temperature",)),
            expr="{koff_ac} / max({kd_ac}, 1e-100)",
        ),
        "pfree": ParamLocalSetting(
            name_setting=NameSetting("pfree", "", TPL),
            expr=f"calc_conc({p_total}, {l_total}, {{kd_ab}}, {{kd_ac}})['pfree']",
        ),
        "l_free": ParamLocalSetting(
            name_setting=NameSetting("l_free", "", TPL),
            expr=f"{l_total} - {p_total} + {{pfree}}",
        ),
        "kab": ParamLocalSetting(
            name_setting=NameSetting("kab", "", TPL),
            expr="{kon_ab} * {l_free}",
        ),
        "kba": ParamLocalSetting(
            name_setting=NameSetting("kba", "", TPL),
            expr="{koff_ab}",
        ),
        "kac": ParamLocalSetting(
            name_setting=NameSetting("kac", "", TPL),
            expr="{kon_ac} * {l_free}",
        ),
        "kca": ParamLocalSetting(
            name_setting=NameSetting("kca", "", TPL),
            expr="{koff_ac}",
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TPL),
            expr="pop_3st({kab}, {kba}, {kac}, {kca}, 0.0, 0.0)['pa']",
        ),
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "", TPL),
            expr="pop_3st({kab}, {kba}, {kac}, {kca}, 0.0, 0.0)['pb']",
        ),
        "pc": ParamLocalSetting(
            name_setting=NameSetting("pc", "", TPL),
            expr="pop_3st({kab}, {kba}, {kac}, {kca}, 0.0, 0.0)['pc']",
        ),
    }


def register() -> None:
    model_factory.register(name=NAME, setting_maker=make_settings_3st_double_binding)
    user_functions = {
        "calc_conc": calculate_concentrations,
        "pop_3st": pop_3st,
    }
    user_function_registry.register(name=NAME, user_functions=user_functions)



from __future__ import annotations

from dataclasses import dataclass
from functools import reduce
from itertools import combinations
from operator import and_
from pathlib import Path
from typing import TYPE_CHECKING

from chemex.parameters import database
from chemex.parameters.name import ParamName

if TYPE_CHECKING:
    from chemex.containers.experiments import Experiments
    from chemex.parameters.setting import Parameters


@dataclass
class Group:
    path: Path
    message: str
    experiments: Experiments


def ids_to_param_name(param_ids: set[str]) -> ParamName:
    if not param_ids:
        return ParamName()
    parameters = database.get_parameters(param_ids)
    param_names = (parameter.param_name for parameter in parameters.values())
    return reduce(and_, param_names)


def group_ids(
    experiments: Experiments,
    parameters: Parameters | None = None,
) -> list[tuple[ParamName, set[str]]]:
    if parameters is None:
        parameters = database.get_parameters(experiments.param_ids)

    ids_vary: set[str] = {
        param_id
        for param_id, param in parameters.items()
        if param.vary and not param.expr
    }

    param_id_sets = (
        set(database.get_parameters(param_id_set))
        for param_id_set in experiments.param_id_sets
    )

    groups: dict[frozenset[str], set[str]] = {
        frozenset(param_id_set & ids_vary): set(param_id_set) & ids_vary
        for param_id_set in param_id_sets
    }

    grouped = True

    while grouped:
        grouped = False
        for (i1, u1), (i2, u2) in combinations(groups.items(), r=2):
            if intersection := frozenset(u1 & u2):
                union = groups.pop(i1) | groups.pop(i2)
                groups.setdefault(intersection, set()).update(union)
                grouped = True
                break

    return sorted((ids_to_param_name(union), union) for union in groups.values())


def create_groups(experiments: Experiments) -> list[Group]:
    """Create groups of datapoints that depend on disjoint sets of variables.

    For example, if the population of the minor state and the exchange
    rate are set to 'fix', chances are that the fit can be decomposed
    residue-specifically.

    """
    id_groups = group_ids(experiments)

    group_name_template = ""
    path_name_template = ""
    group_message_template = ""

    if len(id_groups) > 1:
        group_name_template = "{index}_{pname.folder}"
        path_name_template += "Groups/{group_name}"
        group_message_template = (
            f"Group [magenta]{{group_name}}[/] ({{index}}/{len(id_groups)})"
        )

    groups: list[Group] = []
    for index, (pname, param_ids) in enumerate(id_groups, start=1):
        group_name = group_name_template.format(index=index, pname=pname)
        group_path = Path(path_name_template.format(group_name=group_name))
        group_param_ids = set(database.get_parameters(param_ids))
        group_experiments = experiments.get_relevant_subset(group_param_ids)
        group_message = group_message_template.format(
            index=index,
            group_name=group_name,
        )
        groups.append(Group(group_path, group_message, group_experiments))

    return groups


@dataclass
class ParamTree:
    ids_to_fit: set[str]
    experiments: Experiments
    branches: list[ParamTree]


def create_group_tree(
    experiments: Experiments,
    parameters: Parameters | None = None,
) -> ParamTree:
    if parameters is None:
        parameters = database.get_parameters(experiments.param_ids)

    ids_vary: set[str] = {
        param_id
        for param_id, param in parameters.items()
        if param.vary and not param.expr
    }

    ids_global = {
        param_id
        for param_id in ids_vary
        if len(experiments.get_relevant_subset({param_id})) == len(experiments)
    }

    for param_id in ids_global:
        parameters[param_id].vary = False

    id_groups = group_ids(experiments, parameters)

    branches: list[ParamTree] = []

    if len(id_groups) > 1:
        for _, ids in id_groups:
            sub_experiments = experiments.get_relevant_subset(ids)
            branches.append(create_group_tree(sub_experiments))

    for param_id in ids_global:
        parameters[param_id].vary = True

    return ParamTree(ids_to_fit=ids_global, experiments=experiments, branches=branches)



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np
from scipy.optimize import root

from chemex.models.constraints import pop_3st
from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions
    from chemex.typing import ArrayFloat

NAME = "3st_binding_cs"

TPL = ("temperature", "p_total", "l_total")


def calculate_residuals(
    concentrations: ArrayFloat,
    p_total: float,
    l_total: float,
    kab: float,
    kba: float,
    kd: float,
) -> ArrayFloat:
    p1, p2, pl, l_ = concentrations
    return np.array(
        [
            p_total - p1 - p2 - pl,
            l_total - l_ - pl,
            kab * p1 - kba * p2,
            kd * pl - p2 * l_,
        ],
    )


@lru_cache(maxsize=100)
def calculate_concentrations(
    p_total: float,
    l_total: float,
    kd: float,
    kab: float,
    kba: float,
) -> dict[str, float]:
    concentrations_start = (p_total, 0.0, 0.0, l_total)
    results = root(
        calculate_residuals,
        concentrations_start,
        args=(p_total, l_total, kab, kba, kd),
    )
    p1, p2, pl, l_ = results["x"]
    return {"p1": p1, "p2": p2, "pl": pl, "l": l_}


def make_settings_3st_binding_cs(
    conditions: Conditions,
) -> dict[str, ParamLocalSetting]:
    p0 = conditions.p_total
    l0 = conditions.l_total
    if p0 is None:
        msg = f"'p_total' must be specified to use the '{NAME}' model"
        raise ValueError(msg)
    if l0 is None:
        msg = f"'l_total' must be specified to use the '{NAME}' model"
        raise ValueError(msg)
    return {
        "kab": ParamLocalSetting(
            name_setting=NameSetting("kab", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "kba": ParamLocalSetting(
            name_setting=NameSetting("kba", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "koff": ParamLocalSetting(
            name_setting=NameSetting("koff", "", ("temperature",)),
            value=100.0,
            min=0.0,
            vary=True,
        ),
        "kd": ParamLocalSetting(
            name_setting=NameSetting("kd", "", ("temperature",)),
            value=1e-3,
            min=0.0,
            vary=True,
        ),
        "kon": ParamLocalSetting(
            name_setting=NameSetting("kon", "", ("temperature",)),
            expr="{koff} / max({kd}, 1e-100)",
        ),
        "c_l": ParamLocalSetting(
            name_setting=NameSetting("c_l", "", TPL),
            expr=f"concentrations({p0}, {l0}, {{kab}}, {{kba}}, {{kd}})['l']",
        ),
        "c_p1": ParamLocalSetting(
            name_setting=NameSetting("c_p1", "", TPL),
            expr=f"concentrations({p0}, {l0}, {{kab}}, {{kba}}, {{kd}})['p1']",
        ),
        "c_p2": ParamLocalSetting(
            name_setting=NameSetting("c_p2", "", TPL),
            expr=f"concentrations({p0}, {l0}, {{kab}}, {{kba}}, {{kd}})['p2']",
        ),
        "c_pl": ParamLocalSetting(
            name_setting=NameSetting("c_pl", "", TPL),
            expr=f"concentrations({p0}, {l0}, {{kab}}, {{kba}}, {{kd}})['pl']",
        ),
        "kbc": ParamLocalSetting(
            name_setting=NameSetting("kbc", "", TPL),
            expr="{kon} * {c_l}",
        ),
        "kcb": ParamLocalSetting(
            name_setting=NameSetting("kcb", "", TPL),
            expr="{koff}",
        ),
        "pa": ParamLocalSetting(
            name_setting=NameSetting("pa", "", TPL),
            expr="pop_3st({kab}, {kba}, 0.0, 0.0, {kbc}, {kcb})['pa']",
        ),
        "pb": ParamLocalSetting(
            name_setting=NameSetting("pb", "", TPL),
            expr="pop_3st({kab}, {kba}, 0.0, 0.0, {kbc}, {kcb})['pb']",
        ),
        "pc": ParamLocalSetting(
            name_setting=NameSetting("pc", "", TPL),
            expr="pop_3st({kab}, {kba}, 0.0, 0.0, {kbc}, {kcb})['pc']",
        ),
    }


def register() -> None:
    model_factory.register(name=NAME, setting_maker=make_settings_3st_binding_cs)
    user_functions = {
        "concentrations": calculate_concentrations,
        "pop_3st": pop_3st,
    }
    user_function_registry.register(name=NAME, user_functions=user_functions)



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field
from random import choices
from typing import TYPE_CHECKING, Any, TypeVar

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from chemex.typing import ArrayBool, ArrayFloat

Self = TypeVar("Self", bound="Data")

rng = np.random.default_rng()


def get_scale(exp: ArrayFloat, err: ArrayFloat, calc: ArrayFloat) -> float:
    """Calculate and return the scale factor for the calculated data.

    Args:
        exp (ArrayFloat): The experimental data array.
        err (ArrayFloat): The error associated with the experimental data.
        calc (ArrayFloat): The calculated data array.

    Returns:
        float: The calculated scale factor.

    Raises:
        ZeroDivisionError: If division by zero occurs in the scale calculation.
    """
    if (calc == 0).all():
        return 1.0
    try:
        calc_err2 = calc / err**2
        return sum(exp * calc_err2) / sum(calc * calc_err2)
    except ZeroDivisionError:
        return 1.0


@dataclass
class Data:
    """Dataset with experimental, calculated data and metadata.

    Attributes:
        exp (ArrayFloat): Experimental data.
        err (ArrayFloat): Error in experimental data.
        metadata (NDArray[Any]): Data metadata.
        scale (float): Scale factor for calculated data.
        size (int): Data points count, set post-instantiation.
        calc (ArrayFloat): Calculated data, set post-instantiation.
        mask (ArrayBool): Mask array, set post-instantiation.
        refs (ArrayBool): Reference points array, set post-instantiation.
    """

    exp: ArrayFloat
    err: ArrayFloat
    metadata: NDArray[Any] = field(default_factory=lambda: np.array([]))
    scale: float = 1.0
    size: int = field(init=False)
    calc: ArrayFloat = field(init=False)
    mask: ArrayBool = field(init=False)
    refs: ArrayBool = field(init=False)

    def __post_init__(self):
        """Initialize computed attributes of the Data class."""
        self.size = self.exp.size
        self.calc = np.full_like(self.exp, 1e32)
        self.mask = np.full_like(self.exp, True, dtype=np.bool_)
        self.refs = np.full_like(self.exp, False, dtype=np.bool_)

    def scale_calc(self) -> None:
        """Scale the calculated data based on the experimental data and mask."""
        mask = self.mask
        scale = get_scale(self.exp[mask], self.err[mask], self.calc[mask])
        self.calc *= scale
        self.scale = scale

    def prepare_for_simulation(self) -> None:
        """Prepare the dataset for simulation.

        Set experimental data to calculated values and errors to zero.
        """
        self.exp = self.calc.copy()
        self.err *= 0.0
        if any(self.refs):
            scale = float(100.0 / np.mean(self.exp[self.refs]))
            self.exp *= scale
            self.calc *= scale
            self.scale *= scale

    def monte_carlo(self: Self) -> Self:
        """Generate Data instance for Monte-Carlo simulation.

        Returns:
            Self: New Data instance for Monte-Carlo simulation.
        """
        data = deepcopy(self)
        data.exp = rng.normal(self.calc, self.err)
        return data

    def bootstrap(self: Self) -> Self:
        """Generate Data instance using Bootstrap resampling.

        Returns:
            Self: New Data instance for Bootstrap simulation.
        """
        indexes = np.arange(self.metadata.size)
        pool1 = indexes[self.refs & self.mask]
        pool2 = indexes[~self.refs & self.mask]
        bs_indexes: list[int] = sorted(
            [*choices(pool1, k=pool1.size), *choices(pool2, k=pool2.size)],
        )
        data = deepcopy(self)
        data.metadata[self.mask] = self.metadata[bs_indexes]
        data.exp[self.mask] = self.exp[bs_indexes]
        data.err[self.mask] = self.err[bs_indexes]
        return data

    def sort(self) -> None:
        """Sort the data based on metadata values."""
        sorted_indexes = self.metadata.argsort()
        self.metadata = self.metadata[sorted_indexes]
        self.exp = self.exp[sorted_indexes]
        self.err = self.err[sorted_indexes]
        self.calc = self.calc[sorted_indexes]
        self.refs = self.refs[sorted_indexes]
        self.mask = self.mask[sorted_indexes]

    def any_duplicate(self):
        """Check for duplicate entries in metadata.

        Returns:
            bool: True if duplicates exist, False otherwise.
        """
        return np.unique(self.metadata).size != self.metadata.size

    def __add__(self: Self, other: Self) -> Self:
        """Add two Data instances, combining their experimental data and errors.

        Both Data instances must have the same number of points.

        Args:
            other (Self): Another Data instance.

        Returns:
            Self: New Data instance representing the sum of the two datasets.

        Raises:
            ValueError: If the sizes of the datasets do not match.
        """
        # Check if the sizes of the datasets match
        if self.size != other.size:
            msg = "Sizes of the datasets must be identical to add them."
            raise ValueError(msg)

        # Perform addition if sizes match
        result = deepcopy(self)
        result.exp = self.exp + other.exp
        result.err = np.sqrt(self.err**2 + other.err**2)

        return result



from __future__ import annotations

from functools import lru_cache
from itertools import permutations
from typing import TYPE_CHECKING

import numpy as np
from scipy.constants import constants

from chemex.models.constraints import pop_3st
from chemex.models.factory import model_factory
from chemex.parameters.setting import NameSetting, ParamLocalSetting
from chemex.parameters.userfunctions import user_function_registry

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions
    from chemex.typing import ArrayFloat

NAME = "3st_eyring"

PL = ("p_total", "l_total")
TPL = ("temperature", "p_total", "l_total")


@lru_cache(maxsize=100)
def calculate_kij_3st_eyring(
    dh_b: float,
    ds_b: float,
    dh_c: float,
    ds_c: float,
    dh_ab: float,
    ds_ab: float,
    dh_ac: float,
    ds_ac: float,
    dh_bc: float,
    ds_bc: float,
    temperature: float,
) -> dict[str, float]:
    kelvin = temperature + 273.15
    kbt_h = constants.k * kelvin / constants.h
    rt = constants.R * kelvin
    dh_a = ds_a = 0.0
    ddg_ij = np.array(
        (
            dh_ab - dh_a - kelvin * (ds_ab - ds_a),
            dh_ac - dh_a - kelvin * (ds_ac - ds_a),
            dh_ab - dh_b - kelvin * (ds_ab - ds_b),
            dh_bc - dh_b - kelvin * (ds_bc - ds_b),
            dh_ac - dh_c - kelvin * (ds_ac - ds_c),
            dh_bc - dh_c - kelvin * (ds_bc - ds_c),
        ),
    )
    kij_values: ArrayFloat = kbt_h * np.exp(-ddg_ij / rt)
    kij_values = np.clip(kij_values, 0.0, 1e16)
    kij_names = (f"k{i}{j}" for i, j in permutations("abc", 2))
    return dict(zip(kij_names, kij_values, strict=True))


def create_kij_3st_eyring_settings(temperature: float) -> dict[str, ParamLocalSetting]:
    return {
        f"k{i}{j}": ParamLocalSetting(
            name_setting=NameSetting(f"k{i}{j}", "", TPL),
            expr=(
                f"calculate_kij_3st_eyring({{dh_b}}, {{ds_b}}, {{dh_c}}, {{ds_c}}, "
                f"{{dh_ab}}, {{ds_ab}}, {{dh_ac}}, {{ds_ac}}, {{dh_bc}}, {{ds_bc}},"
                f" {temperature})['k{i}{j}']"
            ),
        )
        for i, j in permutations("abc", 2)
    }


def create_pop_3st_eyring_settings() -> dict[str, ParamLocalSetting]:
    return {
        f"p{state}": ParamLocalSetting(
            name_setting=NameSetting(f"p{state}", "", TPL),
            expr=f"pop_3st({{kab}},{{kba}},{{kac}},{{kca}},{{kbc}},{{kcb}})['p{state}']",
        )
        for state in "abc"
    }


def make_settings_3st_eyring(conditions: Conditions) -> dict[str, ParamLocalSetting]:
    celsius = conditions.temperature
    if celsius is None:
        msg = "The 'temperature' is None"
        raise ValueError(msg)
    return {
        "dh_b": ParamLocalSetting(
            name_setting=NameSetting("dh_b", "", PL),
            value=8e3,
            vary=True,
        ),
        "dh_c": ParamLocalSetting(
            name_setting=NameSetting("dh_c", "", PL),
            value=8e3,
            vary=True,
        ),
        "dh_ab": ParamLocalSetting(
            name_setting=NameSetting("dh_ab", "", PL),
            value=6.5e4,
            vary=True,
        ),
        "dh_bc": ParamLocalSetting(
            name_setting=NameSetting("dh_bc", "", PL),
            value=6.5e4,
            vary=True,
        ),
        "dh_ac": ParamLocalSetting(
            name_setting=NameSetting("dh_ac", "", PL),
            value=1.0e10,
            vary=False,
        ),
        "ds_b": ParamLocalSetting(
            name_setting=NameSetting("ds_b", "", PL),
            value=0.0,
            vary=False,
        ),
        "ds_c": ParamLocalSetting(
            name_setting=NameSetting("ds_c", "", PL),
            value=0.0,
            vary=False,
        ),
        "ds_ab": ParamLocalSetting(
            name_setting=NameSetting("ds_ab", "", PL),
            value=0.0,
            vary=False,
        ),
        "ds_bc": ParamLocalSetting(
            name_setting=NameSetting("ds_bc", "", PL),
            value=0.0,
            vary=False,
        ),
        "ds_ac": ParamLocalSetting(
            name_setting=NameSetting("ds_ac", "", PL),
            value=0.0,
            vary=False,
        ),
        **create_kij_3st_eyring_settings(celsius),
        **create_pop_3st_eyring_settings(),
    }


def register() -> None:
    model_factory.register(name=NAME, setting_maker=make_settings_3st_eyring)
    user_functions = {
        "calculate_kij_3st_eyring": calculate_kij_3st_eyring,
        "pop_3st": pop_3st,
    }
    user_function_registry.register(name=NAME, user_functions=user_functions)



from __future__ import annotations

from functools import total_ordering
from typing import TYPE_CHECKING, Annotated, Literal, TypeVar

from pydantic import (
    BaseModel,
    BeforeValidator,
    Field,
    NonNegativeFloat,
    ValidationError,
    field_validator,
    model_validator,
)

from chemex.configuration.base import to_lower
from chemex.models.model import model

if TYPE_CHECKING:
    from collections.abc import Hashable

Self = TypeVar("Self", bound="Conditions")
T = TypeVar("T")
LabelType = Annotated[Literal["1h", "2h", "13c", "15n"], BeforeValidator(to_lower)]


@total_ordering
class Conditions(BaseModel, frozen=True):
    """Represents experimental conditions for NMR measurements.

    Attributes:
        h_larmor_frq (NonNegativeFloat | None): Larmor frequency of Hydrogen in MHz.
        temperature (float | None): Experimental temperature in degrees Celsius.
        p_total (NonNegativeFloat | None): Total concentration of protein in the sample.
        l_total (NonNegativeFloat | None): Total ligand concentration in the sample.
        d2o (float | None): Fraction of D2O in the solvent, between 0 and 1.
        label (tuple[LabelType, ...]): Tuple of NMR active isotopes used in the experiment.
    """

    h_larmor_frq: NonNegativeFloat | None = None
    temperature: float | None = None
    p_total: NonNegativeFloat | None = None
    l_total: NonNegativeFloat | None = None
    d2o: float | None = Field(gt=0.0, lt=1.0, default=None)
    label: tuple[LabelType, ...] = ()

    def rounded(self: Self) -> Self:
        """Returns a new instance with rounded h_larmor_frq and temperature."""
        h_larmor_frq = round(self.h_larmor_frq, 1) if self.h_larmor_frq else None
        temperature = round(self.temperature, 1) if self.temperature else None
        return self.model_copy(
            update={"h_larmor_frq": h_larmor_frq, "temperature": temperature},
        )

    def match(self: Self, other: Self) -> bool:
        """Checks if the current instance is equivalent to another."""
        return self == self & other

    @property
    def search_keys(self) -> set[Hashable]:
        """Creates a set of hashable search keys based on conditions."""
        return {
            self.h_larmor_frq,
            self.temperature,
            self.p_total,
            self.l_total,
            self.d2o,
        }

    @property
    def section(self) -> str:
        """Generates a string representation of the conditions."""
        parts: list[str] = []
        if self.temperature is not None:
            parts.append(f"T->{self.temperature:.1f}C")
        if self.h_larmor_frq is not None:
            parts.append(f"B0->{self.h_larmor_frq:.1f}MHz")
        if self.p_total is not None:
            parts.append(f"[P]->{self.p_total:e}M")
        if self.l_total is not None:
            parts.append(f"[L]->{self.l_total:e}M")
        if self.d2o is not None:
            parts.append(f"D2O->{self.d2o:.4f}")
        return ", ".join(parts)

    @property
    def folder(self):
        """Generates a folder name representation of the conditions."""
        parts: list[str] = []
        if self.temperature is not None:
            parts.append(f"{self.temperature:.1f}C")
        if self.h_larmor_frq is not None:
            parts.append(f"{self.h_larmor_frq:.1f}MHz")
        if self.p_total is not None:
            parts.append(f"P{self.p_total:e}M")
        if self.l_total is not None:
            parts.append(f"L{self.l_total:e}M")
        if self.d2o is not None:
            parts.append(f"D{self.d2o:.4f}")
        return "_".join(parts)

    @property
    def is_deuterated(self) -> bool:
        """Checks if the conditions include deuterium."""
        return "2h" in self.label

    def select_conditions(self: Self, conditions_selection: tuple[str, ...]) -> Self:
        """Selects specific conditions based on given keys.

        Args:
            conditions_selection (tuple[str, ...]): Keys to select conditions by.

        Returns:
            A new instance of Conditions with selected conditions.
        """
        return type(self).model_construct(
            **{
                key: value
                for key, value in self.model_dump().items()
                if key in conditions_selection
            },
        )

    def __and__(self: Self, other: object) -> Self:
        if not isinstance(other, type(self)):
            return NotImplemented
        """Defines bitwise AND operation for Conditions instances."""
        self_dict = self.model_dump()
        other_dict = other.model_dump()
        intersection = {
            key: value for key, value in self_dict.items() if other_dict[key] == value
        }
        return type(self).model_construct(**intersection)

    def __lt__(self, other: object) -> bool:
        """Defines less than operation for Conditions instances."""
        if not isinstance(other, type(self)):
            return NotImplemented
        tuple_self = tuple(
            value if value is not None else -1e16
            for value in self.model_dump().values()
        )
        tuple_other = tuple(
            value if value is not None else -1e16
            for value in other.model_dump().values()
        )
        return tuple_self < tuple_other


@total_ordering
class ConditionsFromFile(Conditions, frozen=True):
    """Extends Conditions to support loading experimental conditions from a file.

    This class includes additional validations specific to the file-based input.
    """

    @model_validator(mode="before")
    def key_to_lower(cls, model: dict[str, T]) -> dict[str, T]:
        """Converts keys of the model dictionary to lowercase."""
        return {to_lower(k): v for k, v in model.items()}

    @field_validator("d2o")
    def validate_d2o(cls, d2o: float | None) -> float | None:
        """Validates the d2o field for specific model requirements."""
        if "hd" in model.name and d2o is None:
            msg = 'To use the "hd" model, d2o must be provided'
            raise ValidationError(msg)
        return d2o

    @field_validator("temperature")
    def validate_temperature(cls, temperature: float | None) -> float | None:
        """Validates temperature field for specific model requirements."""
        if "eyring" in model.name and temperature is None:
            msg = 'To use the "eyring" model, "temperature" must be provided'
            raise ValidationError(msg)
        return temperature

    @model_validator(mode="after")
    def validate_p_total_l_total(self: Self) -> Self:
        """Validates both p_total and l_total for specific model requirements."""
        are_not_both_set = self.p_total is None or self.l_total is None
        if "binding" in model.name and are_not_both_set:
            msg = 'To use the "binding" model, "p_total" and "l_total" must be provided'
            raise ValidationError(msg)
        return self



"""Utility module for list and string manipulation in Pydantic models.

This module provides utility functions for ensuring a variable is in list format and
for converting strings to lowercase. It also includes a Pydantic BaseModel subclass
which applies these utilities to convert all keys in a model to lowercase.

Typical usage example:

  variable = "Sample"
  variable_list = ensure_list(variable)
  lower_variable = to_lower("HELLO")
  model_instance = BaseModelLowerCase.parse_obj({"Name": "Alice"})
"""
from __future__ import annotations

from typing import TypeVar, cast

from pydantic import BaseModel, model_validator

T = TypeVar("T")


def ensure_list(variable: T | list[T] | None) -> T | list[T]:
    """Ensures that the input variable is returned as a list.

    If the input variable is already a list, it is returned as-is.
    If the variable is None, an empty list is returned.
    Otherwise, the variable is wrapped in a list and returned.

    Parameters:
    variable (T | list[T] | None): The variable to be ensured as a list.

    Returns:
    T | list[T]: The input variable as a list.
    """
    if isinstance(variable, list):
        return cast(list[T], variable)
    if variable is None:
        return []
    return [variable]


def to_lower(string: T) -> T:
    """Converts a string to lowercase if it is of type str.

    If the input is not a string, it is returned unchanged.

    Parameters:
    string (T): The string to be converted to lowercase.

    Returns:
    T: The lowercase string if input is a string; otherwise, the original input.
    """
    if isinstance(string, str):
        return string.lower()
    return string


class BaseModelLowerCase(BaseModel):
    """A Pydantic BaseModel class that converts all keys in the model to lowercase.

    This is achieved using a model validator that operates before model initialization.
    The validator applies the `to_lower` function to each key of the model.
    """

    @model_validator(mode="before")
    def key_to_lower(cls, model: dict[str, T]) -> dict[str, T]:
        """Model validator to convert all dictionary keys to lowercase.

        Parameters:
        model (dict[str, T]): The dictionary model with string keys.

        Returns:
        dict[str, T]: The modified dictionary with all keys in lowercase.
        """
        return {to_lower(k): v for k, v in model.items()}



"""Factories for creating different parts of an experiment."""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar

if TYPE_CHECKING:
    from collections.abc import Callable, MutableMapping
    from pathlib import Path

    from chemex.configuration.experiment import ExperimentConfig
    from chemex.containers.data import Data
    from chemex.containers.profile import Filterer, PulseSequence
    from chemex.nmr.spectrometer import Spectrometer
    from chemex.parameters.spin_system import SpinSystem
    from chemex.plotters.plotter import Plotter
    from chemex.printers.data import Printer

    Dataset = list[tuple[SpinSystem, Data]]
    ConfigType = ExperimentConfig[Any, Any]

    ConfigCreator = Callable[..., ConfigType]
    PropagatorCreator = Callable[..., Spectrometer]
    SequenceCreator = Callable[..., PulseSequence]
    DatasetCreator = Callable[[Path, ConfigType], Dataset]
    FiltererCreator = Callable[..., Filterer]
    PrinterCreator = Callable[[], Printer]
    PlotterCreator = Callable[..., Plotter]


@dataclass
class Creators:
    config_creator: ConfigCreator
    spectrometer_creator: PropagatorCreator
    sequence_creator: SequenceCreator
    dataset_creator: DatasetCreator
    filterer_creator: FiltererCreator
    printer_creator: PrinterCreator
    plotter_creator: PlotterCreator

    def create_config(self, config_dict: MutableMapping[str, Any]) -> ConfigType:
        """Create a configuration object of a specific type."""
        return self.config_creator(**config_dict)

    def create_spectrometer(
        self,
        config: ConfigType,
        spin_system: SpinSystem,
    ) -> Spectrometer:
        """Create and initialize a spectrometer of a specific type."""
        return self.spectrometer_creator(config, spin_system)

    def create_sequence(self, config: ConfigType) -> PulseSequence:
        """Create a sequence of a specific type."""
        return self.sequence_creator(config.experiment)

    def create_dataset(self, base_path: Path, settings: ConfigType) -> Dataset:
        """Create a dataset of a specific type."""
        return self.dataset_creator(base_path, settings)

    def create_filterer(
        self,
        config: ConfigType,
        spectrometer: Spectrometer,
    ) -> Filterer:
        """Create a filterer used to remove undesired data points."""
        return self.filterer_creator(config=config, spectrometer=spectrometer)

    def create_printer(self) -> Printer:
        """Create a filterer used to remove undesired data points."""
        return self.printer_creator()

    def create_plotter(self, filename: Path, config: ConfigType) -> Plotter:
        """Create a filterer used to remove undesired data points."""
        return self.plotter_creator(filename=filename, config=config)


class Factories:
    """Factory for creating all parts of an experiment."""

    creators_registry: ClassVar[dict[str, Creators]] = {}

    def register(self, type: str, creators: Creators):
        """Register a new propagtor type."""
        self.creators_registry[type] = creators

    def get(self, type: str) -> Creators:
        try:
            return self.creators_registry[type]
        except KeyError:
            msg = f"Unknown  type {type!r}"
            raise ValueError(msg) from None


factories = Factories()



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

from chemex.models.model import model
from chemex.nmr.constants import J_COUPLINGS
from chemex.nmr.rates import get_model_free_expressions
from chemex.parameters.setting import LocalSettings, NameSetting, ParamLocalSetting

if TYPE_CHECKING:
    from chemex.configuration.conditions import Conditions
    from chemex.nmr.basis import Basis


def _update_expr_for_proton_exchange(
    settings: dict[str, ParamLocalSetting],
    state: str,
    basis: Basis,
) -> None:
    # As 1HN sites exchange with water, the "expr" should be modified accordingly
    # for spin system including an amide proton
    if basis.spin_system == "hn":
        expr = f"{{r2_s_{state}}} + {{r1a_is_{state}}} - {{r1_s_{state}}}"
        settings[f"r2a_s_{state}"].expr = expr
    elif basis.spin_system == "nh":
        expr = f"{{r2_i_{state}}} + {{r1a_is_{state}}} - {{r1_i_{state}}}"
        settings[f"r2a_i_{state}"].expr = expr
        settings[f"r1a_is_{state}"].expr = ""


def _add_temp_coef_param_settings(
    settings: dict[str, ParamLocalSetting],
    state: str,
    conditions: Conditions,
) -> None:
    settings[f"dwm_i_a{state}"] = ParamLocalSetting(
        name_setting=NameSetting(f"dwm_a{state}", "i"),
        value=0.0,
        min=-1.0,
        max=1.0,
        vary=True,
    )
    settings[f"dwp_i_a{state}"] = ParamLocalSetting(
        name_setting=NameSetting(f"dwp_a{state}", "i"),
        value=0.0,
        min=-100.0,
        max=100.0,
        vary=True,
    )
    settings[f"dwm_s_a{state}"] = ParamLocalSetting(
        name_setting=NameSetting(f"dwm_a{state}", "s"),
        value=0.0,
        min=-1.0,
        max=1.0,
        vary=True,
    )
    settings[f"dwp_s_a{state}"] = ParamLocalSetting(
        name_setting=NameSetting(f"dwp_a{state}", "s"),
        value=0.0,
        min=-100.0,
        max=100.0,
        vary=True,
    )
    temp = conditions.temperature
    setting_dw_i = settings[f"dw_i_a{state}"]
    setting_dw_i.vary = False
    setting_dw_i.expr = f"{{dwp_i_a{state}}} + {temp} * {{dwm_i_a{state}}}"
    setting_dw_s = settings[f"dw_s_a{state}"]
    setting_dw_s.vary = False
    setting_dw_s.expr = f"{{dwp_s_a{state}}} + {temp} * {{dwm_s_a{state}}}"


def _add_dw_param_settings(
    settings: dict[str, ParamLocalSetting],
    state: str,
    conditions: Conditions,
) -> None:
    settings[f"dw_i_a{state}"] = ParamLocalSetting(
        name_setting=NameSetting(f"dw_a{state}", "i", ("temperature",)),
        value=0.0,
        min=-100.0,
        max=100.0,
        vary=True,
    )
    settings[f"dw_s_a{state}"] = ParamLocalSetting(
        name_setting=NameSetting(f"dw_a{state}", "s", ("temperature",)),
        value=0.0,
        min=-100.0,
        max=100.0,
        vary=True,
    )
    settings[f"cs_i_{state}"].expr = f"{{cs_i_a}} + {{dw_i_a{state}}}"
    settings[f"cs_s_{state}"].expr = f"{{cs_s_a}} + {{dw_s_a{state}}}"

    if model.temp_coef:
        _add_temp_coef_param_settings(settings, state, conditions)


def _set_equal_to_a(settings: dict[str, ParamLocalSetting]) -> None:
    for name, setting in settings.items():
        if not setting.expr:
            setting.expr = f"{{{name[:-1]}a}}"


def create_base_param_settings(
    basis: Basis,
    state: str,
    conditions: Conditions,
) -> dict[str, ParamLocalSetting]:
    ext = basis.extension
    settings = {
        f"tauc_{state}": ParamLocalSetting(
            name_setting=NameSetting(f"tauc_{state}", "g", ("temperature",)),
            value=4.0,
            min=0.0,
        ),
        f"s2_{state}": ParamLocalSetting(
            name_setting=NameSetting(f"s2_{state}", "g", ("temperature",)),
            value=0.9,
            min=0.0,
        ),
        f"khh_{state}": ParamLocalSetting(
            name_setting=NameSetting(f"khh_{state}", "g", ("temperature",)),
            value=0.0,
        ),
        f"r2_i_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"r2{ext}_{state}",
                "i",
                ("temperature", "h_larmor_frq"),
            ),
            value=10.0,
            min=0.0,
        ),
        f"r2_s_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"r2_{state}",
                "s",
                ("temperature", "h_larmor_frq"),
            ),
            value=10.0,
            min=0.0,
        ),
        f"r1_i_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"r1{ext}_{state}",
                "i",
                ("temperature", "h_larmor_frq"),
            ),
            value=1.5,
            min=0.0,
        ),
        f"r1_s_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"r1_{state}",
                "s",
                ("temperature", "h_larmor_frq"),
            ),
            value=1.5,
            min=0.0,
        ),
        f"r2a_i_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"r2a{ext}_{state}",
                "i",
                ("temperature", "h_larmor_frq"),
            ),
            value=12.0,
            min=0.0,
            expr=f"{{r2_i_{state}}} - {{r1_s_{state}}}",
        ),
        f"r2a_s_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"r2a_{state}",
                "s",
                ("temperature", "h_larmor_frq"),
            ),
            value=12.0,
            min=0.0,
            expr=f"{{r2_s_{state}}} - {{r1_i_{state}}}",
        ),
        f"r2mq_is_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"r2mq_{state}",
                "is",
                ("temperature", "h_larmor_frq"),
            ),
            value=15.0,
            min=0.0,
        ),
        f"r1a_is_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"r1a{ext}_{state}",
                "is",
                ("temperature", "h_larmor_frq"),
            ),
            value=4.0,
            min=0.0,
            expr=f"{{r1_i_{state}}} + {{r1_s_{state}}}",
        ),
        f"etaxy_i_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"etaxy{ext}_{state}",
                "i",
                ("temperature", "h_larmor_frq"),
            ),
            value=0.0,
        ),
        f"etaxy_s_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"etaxy_{state}",
                "s",
                ("temperature", "h_larmor_frq"),
            ),
            value=0.0,
        ),
        f"etaz_i_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"etaz{ext}_{state}",
                "i",
                ("temperature", "h_larmor_frq"),
            ),
            value=0.0,
        ),
        f"etaz_s_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"etaz_{state}",
                "s",
                ("temperature", "h_larmor_frq"),
            ),
            value=0.0,
        ),
        f"sigma_is_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"sigma_{state}",
                "is",
                ("temperature", "h_larmor_frq"),
            ),
            value=0.0,
        ),
        f"mu_is_{state}": ParamLocalSetting(
            name_setting=NameSetting(
                f"mu_{state}",
                "is",
                ("temperature", "h_larmor_frq"),
            ),
            value=0.0,
        ),
        f"cs_i_{state}": ParamLocalSetting(
            name_setting=NameSetting(f"cs_{state}", "i", ("temperature",)),
            value=0.0,
        ),
        f"cs_s_{state}": ParamLocalSetting(
            name_setting=NameSetting(f"cs_{state}", "s", ("temperature",)),
            value=0.0,
        ),
        f"j_is_{state}": ParamLocalSetting(
            name_setting=NameSetting(f"j_{state}", "is"),
            value=J_COUPLINGS.get(basis.spin_system, 0.0),
        ),
        f"d_{state}": ParamLocalSetting(
            name_setting=NameSetting(f"d_{state}", "", ("temperature",)),
            value=0.0,
        ),
    }

    _update_expr_for_proton_exchange(settings, state, basis)

    if state != "a":
        _set_equal_to_a(settings)
        _add_dw_param_settings(settings, state, conditions)

    return settings


def _build_model_free_settings(
    settings: LocalSettings,
    basis: Basis,
    conditions: Conditions,
) -> LocalSettings:
    model_free_expressions = get_model_free_expressions(basis, conditions)
    settings_mf = deepcopy(settings)
    for name, setting in settings_mf.items():
        if name in model_free_expressions:
            setting.expr = model_free_expressions[name]

    return settings_mf


def _select_relevant_settings(
    all_settings: LocalSettings,
    basis: Basis,
) -> LocalSettings:
    pool = set(all_settings) & set(basis.matrices)
    selection: set[str] = set()
    while pool:
        name = pool.pop()
        selection.add(name)
        pool |= all_settings[name].dependencies

    return {name: all_settings[name] for name in selection}


def build_spin_param_settings(
    basis: Basis,
    conditions: Conditions,
) -> tuple[LocalSettings, LocalSettings]:
    all_settings: LocalSettings = {}
    for state in model.states:
        all_settings.update(create_base_param_settings(basis, state, conditions))

    all_settings_mf = _build_model_free_settings(all_settings, basis, conditions)

    settings = _select_relevant_settings(all_settings, basis)
    settings_mf = _select_relevant_settings(all_settings_mf, basis)

    # Add back to `settings_mf` the parameters that were selected in `settings`,
    # but not in settings_mf. This can happen when a constraint require parameters
    # that are not in the basis. This way the starting value of these parameters
    # can be calculated using the model free parameters.
    if not model.model_free:
        names = set(settings) - set(settings_mf)
        settings_mf |= {name: all_settings_mf[name] for name in names}

    return settings, settings_mf
